# pyOCD debugger
# Copyright (c) 2006-2013 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...debug.svd.loader import SVDFile
from ..family.target_nRF52 import NRF52

FLASH_ALGO = { 'load_address' : 0x20000000,
               'instructions' : [
                                0xE00ABE00, 0x062D780D, 0x24084068, 0xD3000040, 0x1E644058, 0x1C49D1FA, 0x2A001E52, 0x4770D1F2,
                                0x47702000, 0x47702000, 0x4c26b570, 0x60602002, 0x60e02001, 0x68284d24, 0xd00207c0, 0x60602000,
                                0xf000bd70, 0xe7f6f82c, 0x4c1eb570, 0x60612102, 0x4288491e, 0x2001d302, 0xe0006160, 0x4d1a60a0,
                                0xf81df000, 0x07c06828, 0x2000d0fa, 0xbd706060, 0x4605b5f8, 0x4813088e, 0x46142101, 0x4f126041,
                                0xc501cc01, 0x07c06838, 0x1e76d006, 0x480dd1f8, 0x60412100, 0xbdf84608, 0xf801f000, 0x480ce7f2,
                                0x06006840, 0xd00b0e00, 0x6849490a, 0xd0072900, 0x4a0a4909, 0xd00007c3, 0x1d09600a, 0xd1f90840,
                                0x00004770, 0x4001e500, 0x4001e400, 0x10001000, 0x40010400, 0x40010500, 0x40010600, 0x6e524635,
                                0x00000000, ],
               'pc_init'          : 0x20000021,
               'pc_eraseAll'      : 0x20000029,
               'pc_erase_sector'  : 0x20000049,
               'pc_program_page'  : 0x20000071,
               'begin_data'       : 0x20002000, # Analyzer uses a max of 0.5 KB data (128 pages * 4 bytes / page)
               'page_buffers'    : [0x20002000, 0x20003000],   # Enable double buffering
               'begin_stack'      : 0x20001000,
               'static_base'      : 0x20000170,
               'min_program_length' : 4,
               'analyzer_supported' : True,
               'analyzer_address' : 0x20004000  # Analyzer 0x20004000..0x20004600
              }

class NRF52840(NRF52):

    MEMORY_MAP = MemoryMap(
        FlashRegion(    start=0x0,         length=0x100000,     blocksize=0x1000, is_boot_memory=True,
            algo=FLASH_ALGO),
        # User Information Configation Registers (UICR) as a flash region
        FlashRegion(    start=0x10001000,  length=0x400,        blocksize=0x400, is_testable=False,
            algo=FLASH_ALGO),
        RamRegion(      start=0x20000000,  length=0x40000)
        )

    def __init__(self, session):
        super(NRF52840, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("nrf52840.svd")



# pyOCD debugger
# Copyright (c) 2021 mikisama
# Copyright (C) 2021 Ciro Cattuto <ciro.cattuto@gmail.com>
# Copyright (C) 2021 Simon D. Levy <simon.d.levy@gmail.com>
# Copyright (C) 2022 Johan Carlsson <johan.carlsson@teenage.engineering>
# Copyright (c) 2022 Samuel Dewan
# Copyright (C) 2023 Tejaswini Dasika <tejaswinidasika@gmail.com>
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from ctypes import Structure, c_char, c_int32, c_uint32, sizeof
import struct
from typing import Optional, Sequence

from ..core.memory_map import MemoryMap, MemoryRegion, MemoryType
from ..core.soc_target import SoCTarget
from ..core import exceptions


class SEGGER_RTT_BUFFER_UP(Structure):
    """@brief `SEGGER RTT Ring Buffer` target to host."""

    _fields_ = [
        ("sName", c_uint32),
        ("pBuffer", c_uint32),
        ("SizeOfBuffer", c_uint32),
        ("WrOff", c_uint32),
        ("RdOff", c_uint32),
        ("Flags", c_uint32),
    ]

class SEGGER_RTT_BUFFER_DOWN(Structure):
    """@brief `SEGGER RTT Ring Buffer` host to target."""

    _fields_ = [
        ("sName", c_uint32),
        ("pBuffer", c_uint32),
        ("SizeOfBuffer", c_uint32),
        ("WrOff", c_uint32),
        ("RdOff", c_uint32),
        ("Flags", c_uint32),
    ]

class SEGGER_RTT_CB(Structure):
    """@brief `SEGGER RTT control block` structure. """

    _fields_ = [
        ("acID", c_char * 16),
        ("MaxNumUpBuffers", c_int32),
        ("MaxNumDownBuffers", c_int32)
    ]



class RTTUpChannel(ABC):
    """@brief Wrapper for an RTT up channel for target to host data transfer. """

    name: Optional[str]
    size: int

    @abstractmethod
    def read(self) -> bytes:
        """@brief Read all available data from RTT channel. """


class RTTDownChannel(ABC):
    """@brief Wrapper for an RTT down channel for host to target data transfer. """

    name: Optional[str]
    size: int

    @abstractmethod
    def write(self, data: bytes, blocking = False) -> int:
        """@brief Write data to RTT channel.

        Write as much of the provided data as possible to an the RTT down
        channel.

        @param data The data to be written.
        @param blocking If true will block until all data is sent.

        @return The number of bytes written to the target.
        """


class RTTControlBlock(ABC):
    """@brief Represents an RTT control block.

    This helper class can be used to search for and parse an RTT control block.
    Once the control block is found and parsed (using the start() method)
    RTTUpChannel and RTTDownChannel objects will be created to facilitate
    communication with the target.
    """

    up_channels: Sequence[RTTUpChannel]
    down_channels: Sequence[RTTDownChannel]

    @abstractmethod
    def start(self):
        """@brief Find the RTT control block on the target.

        Searches for the RTT control block. Once found, the up_channels and
        down_channels lists are populated with RTTUpChannel and RTTDownChannel
        objects that can be used to communicate with the target.
        """
        pass

    @classmethod
    def from_target(cls, target: SoCTarget, address: int = None,
                    size: int = None, control_block_id: bytes = b'SEGGER RTT'):
        """@brief Create an RTTControlBlock object using a given target.

        This function creates an instance of an appropriate RTTControlBlock
        subclass depending on the provided target.

        @param target The target with which RTT communication is desired.
        @param address Base address for control block search range.
        @param size Control block search range. If 0 the control block will be
                    expected to be located at the provided address.
        @param control_block_id The control block ID string to search for. Must
                                be at most 16 bytes long.  Will be padded with
                                zeroes if less than 16 bytes.

        @return An instance of an appropriate RTTControlBlock subclass.
        """
        # TODO: Handle targets connected with jlink differently
        return GenericRTTControlBlock(target, address = address, size = size,
                                      control_block_id = control_block_id)




class GenericRTTUpChannel(RTTUpChannel):
    """@brief Software implementation of RTT up channel. Does not require any
              support from interface.
    """

    _target: SoCTarget
    name: Optional[str]
    _buffer_address: int
    size: int
    _offsets_addr: int
    _desc_addr: int

    def __init__(self, target: SoCTarget, desc_addr: int):
        """
        @param target Target to communicate with.
        @param desc_addr Address of up buffer descriptor.
        """
        self._target = target
        self._desc_addr = desc_addr
        self._offsets_addr = self._desc_addr + SEGGER_RTT_BUFFER_UP.WrOff.offset

        self._read_descriptor()

    def _read_descriptor(self):
        # Get buffer descriptor
        up_buffer_words = sizeof(SEGGER_RTT_BUFFER_UP) // 4
        data = self._target.read_memory_block32(self._desc_addr, up_buffer_words)
        descriptor: SEGGER_RTT_BUFFER_UP = SEGGER_RTT_BUFFER_UP(*data)

        # Get name if there is one
        if descriptor.sName != 0:
            data = b''
            while True:
                data += bytes(self._target.read_memory_block8(descriptor.sName, 32))
                name_length = data.find(b'\0')
                if name_length != -1:
                    self.name = data[:name_length].decode("utf-8", "backslashreplace")
                    break
                elif len(data) >= 512:
                    # Give up, this probably isn't a valid string
                    self.name = data.decode("utf-8", "backslashreplace")
                    break
        else:
            self.name = None

        self._buffer_address = descriptor.pBuffer
        self.size = descriptor.SizeOfBuffer

    @property
    def bytes_available(self) -> int:
        """@brief Number of bytes available to be read from up channel. """
        if (self.size == 0) or (self._buffer_address == 0):
            # descriptor is not yet populated
            self._read_descriptor()
            if (self.size == 0) or (self._buffer_address == 0):
                # descriptor is still not populated
                return 0

        # Get offsets
        write_off, read_off = self.target.read_memory_block32(self._offsets_addr, 2)

        if (write_off >= self.size) or (read_off >= self.size):
            raise exceptions.RTTError("Invalid up buffer")
        elif write_off == self.read_off:
            return 0
        elif write_off > read_off:
            return write_off - read_off
        else:
            return (self.size - read_off) + write_off

    def read(self) -> bytes:
        """@brief Read all available data from RTT channel. """
        if (self.size == 0) or (self._buffer_address == 0):
            # descriptor is not yet populated
            self._read_descriptor()
            if (self.size == 0) or (self._buffer_address == 0):
                # descriptor is still not populated
                return b''

        # Get offsets
        write_off, read_off = self._target.read_memory_block32(self._offsets_addr, 2)

        if (write_off >= self.size) or (read_off >= self.size):
            raise exceptions.RTTError("Invalid up buffer")
        elif write_off == read_off:
            # empty
            return b''
        elif write_off > read_off:
            """
            |oooooo|xxxxxxxxxxxx|oooooo|
            0    rdOff        WrOff    SizeOfBuffer
            """
            data = self._target.read_memory_block8(self._buffer_address + read_off,
                                                   write_off - read_off)
        else:
            """
            |xxxxxx|oooooooooooo|xxxxxx|
            0    WrOff        RdOff    SizeOfBuffer
            """
            data = self._target.read_memory_block8(self._buffer_address + read_off,
                                                   self.size - read_off)
            data += self._target.read_memory_block8(self._buffer_address, write_off)

        # Update read offset
        self._target.write32(self._offsets_addr + 4, write_off)
        return bytes(data)


class GenericRTTDownChannel(RTTDownChannel):
    """@brief Software implementation of RTT down channel. Does not require any
              support from interface.
    """

    _target: SoCTarget
    name: Optional[str]
    _buffer_address: int
    size: int
    _offsets_addr: int
    _desc_addr: int

    def __init__(self, target: SoCTarget, desc_addr: int):
        """
        @param target Target to communicate with.
        @param desc_addr Address of down buffer descriptor.
        """
        self._target = target
        self._desc_addr = desc_addr
        self._offsets_addr = self._desc_addr + SEGGER_RTT_BUFFER_DOWN.WrOff.offset

        self._read_descriptor()

    def _read_descriptor(self):
        # Get buffer descriptor
        up_buffer_words = sizeof(SEGGER_RTT_BUFFER_DOWN) // 4
        data = self._target.read_memory_block32(self._desc_addr, up_buffer_words)
        descriptor: SEGGER_RTT_BUFFER_DOWN = SEGGER_RTT_BUFFER_DOWN(*data)

        # Get name if there is one
        if descriptor.sName != 0:
            data = b''
            while True:
                data += bytes(self._target.read_memory_block8(descriptor.sName, 64))
                name_length = data.find(b'\0')
                if name_length != -1:
                    self.name = data[:name_length].decode("utf-8", "backslashreplace")
                    break
                elif len(data) > 512:
                    # Give up, this probably isn't a valid string
                    self.name = data.decode("utf-8", "backslashreplace")
                    break
        else:
            self.name = None

        self._buffer_address = descriptor.pBuffer
        self.size = descriptor.SizeOfBuffer

    @property
    def bytes_free(self) -> int:
        """@brief Number of bytes free in RTT down channel ring buffer."""
        if (self.size == 0) or (self._buffer_address == 0):
            # descriptor is not yet populated
            self._read_descriptor()
            if (self.size == 0) or (self._buffer_address == 0):
                # descriptor is still not populated
                return 0

        # Get offsets
        write_off, read_off = self._target.read_memory_block32(self._offsets_addr, 2)

        if (write_off >= self.size) or (read_off >= self.size):
            raise exceptions.RTTError("Invalid down buffer")
        elif write_off == self.read_off:
            return self.size
        elif write_off > read_off:
            return (self.size - write_off) + (read_off - 1)
        else:
            return read_off - write_off - 1

    def write(self, data: bytes, blocking = False) -> int:
        """@brief Write data to RTT channel.

        Write as much of the provided data as possible to an the RTT down
        channel.

        @param data The data to be written.
        @param blocking If true will block until all data is sent.

        @return The number of bytes written to the target.
        """
        if (self.size == 0) or (self._buffer_address == 0):
            # descriptor is not yet populated
            self._read_descriptor()
            if (self.size == 0) or (self._buffer_address == 0):
                # descriptor is still not populated
                return 0

        if blocking:
            # Call non-blocking version until all data is written
            while data:
                bytes_sent: int = self.write(data, blocking = False)
                data = data[bytes_sent:]
            return

        # Get offsets
        write_off, read_off = self._target.read_memory_block32(self._offsets_addr, 2)

        if (write_off >= self.size) or (read_off >= self.size):
            raise exceptions.RTTError("Invalid down buffer")


        bytes_written: int = 0
        if write_off >= read_off:
            # There is some space to fill at the top of the buffer
            free_space: int = self.size - write_off
            if read_off == 0:
                # Can't use the last element in the buffer
                free_space -= 1
            data_to_write: bytes = data[:free_space]
            self._target.write_memory_block8(self._buffer_address + write_off,
                                             data_to_write)
            bytes_written = len(data_to_write)
            data = data[bytes_written:]
            write_off = (write_off + bytes_written) % self.size

        free_space: int = read_off - write_off - 1
        if free_space < 0:
            free_space = 0

        bytes_to_write: int = min(free_space, len(data))
        self._target.write_memory_block8(self._buffer_address + write_off,
                                         data[:bytes_to_write])
        bytes_written += bytes_to_write
        write_off += bytes_to_write

        # Store new write offset
        self._target.write32(self._offsets_addr, write_off)
        return bytes_written


class GenericRTTControlBlock(RTTControlBlock):
    """@brief Software implementation of RTT control block helper. Does not
              require any support from interface.
    """

    target: SoCTarget
    _cb_search_address: int
    _cb_search_size_bytes: int
    _control_block_id: Sequence[int]

    def __init__(self, target: SoCTarget, address: int = None,
                 size: int = None, control_block_id: bytes = b'SEGGER RTT'):
        """
        @param target The target with which RTT communication is desired.
        @param address Base address for control block search range.
        @param size Control block search range. If 0 the control block will be
                    expected to be located at the provided address.
        @param control_block_id The control block ID string to search for. Must
                                be at most 16 bytes long.  Will be padded with
                                zeroes if less than 16 bytes.
        """
        self.target = target
        self.up_channels = list()
        self.down_channels = list()

        if address is None:
            memory_map: MemoryMap = self.target.get_memory_map()
            ram_region: MemoryRegion = memory_map.get_default_region_of_type(MemoryType.RAM)

            self._cb_search_address = ram_region.start
            if size is None:
                size = ram_region.length
        else:
            self._cb_search_address = address

        if size is None:
            # Address was specified, but size was not. Assume that the control
            # block is located exactly at the provided address.
            self._cb_search_size_bytes = 0
        else:
            self._cb_search_size_bytes = size
        self._control_block_id = control_block_id

    def _find_control_block(self) -> Optional[int]:
        addr: int = self._cb_search_address & ~0x3
        search_size: int  = self._cb_search_size_bytes
        if search_size < len(self._control_block_id):
            search_size = len(self._control_block_id)

        id_len = len(self._control_block_id)
        offset: int = 0

        while search_size:
            read_size = min(search_size, 32)
            data = self.target.read_memory_block8(addr, read_size)

            if not data:
                break

            for byte in data:
                if byte == self._control_block_id[offset]:
                    offset += 1
                    if offset == id_len:
                        break
                else:
                    num_skip_words = (offset + 1)
                    addr += (num_skip_words * 1)
                    search_size -= num_skip_words
                    offset = 0

            if offset == id_len:
                break

        return addr if offset == id_len else None

    def start(self):
        """@brief Find the RTT control block on the target.

        Searches for the RTT control block. Once found, the up_channels and
        down_channels lists are populated with RTTUpChannel and RTTDownChannel
        objects that can be used to communicate with the target.
        """

        cb_addr = self._find_control_block()

        if cb_addr is None:
            raise exceptions.RTTError("Control block not found")

        # Get control block info
        num_up_buffs = self.target.read32(cb_addr + SEGGER_RTT_CB.MaxNumUpBuffers.offset)
        num_down_buffs = self.target.read32(cb_addr + SEGGER_RTT_CB.MaxNumDownBuffers.offset)

        # Setup up channels
        up_base = cb_addr + sizeof(SEGGER_RTT_CB)
        for i in range(num_up_buffs):
            addr = up_base + (i * sizeof(SEGGER_RTT_BUFFER_UP))
            self.up_channels.append(GenericRTTUpChannel(self.target, addr))

        # Setup down channels
        down_base = up_base + num_up_buffs * sizeof(SEGGER_RTT_BUFFER_UP)
        for i in range(num_down_buffs):
            addr = down_base + (i * sizeof(SEGGER_RTT_BUFFER_DOWN))
            self.down_channels.append(GenericRTTDownChannel(self.target, addr))



# pyOCD debugger
# Copyright (c) 2023 Protech Engineering
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .component import CoreSightComponent

LOG = logging.getLogger(__name__)

class TraceFunnel(CoreSightComponent):
    """@brief CoreSight Trace Funnel"""

    # Register definitions.
    #
    # The addresses are offsets from the base address.
    CSTF = 0x00000000
    CSTF_ENSX_MASK = 0xFF

    DEVID = 0x00000FC8
    DEVID_PORTCOUNT_MASK = 0xF

    def __init__(self, ap, cmpid=None, addr=None):
        """@brief Standard CoreSight component constructor."""
        super().__init__(ap, cmpid, addr)
        self._available_channels = 2

    @property
    def available_channels(self) -> int:
        """@brief Number of input ports connected to the funnel"""
        return self._available_channels

    def init(self) -> None:
        """@brief Reads Funnel connected channels and enables them all by default."""
        devid = self.ap.read32(self.address + TraceFunnel.DEVID)
        self._available_channels = devid & TraceFunnel.DEVID_PORTCOUNT_MASK
        self.enable()

    def set_enabled_channels(self, channels: int) -> bool:
        """@brief Sets the enabled Trace Funnel channels.

        @param channels Word describing the desired state for the funnel channels.
            Setting the n-th bit of this word high enables the corresponding n-th channel, setting it low disables it.
        """
        valid_channels_mask = 2**self.available_channels-1
        if channels & ~valid_channels_mask:
            LOG.warning(f"Trace Funnel: Trying to enable too many channels. Only {self.available_channels} channels are present")
            return False

        cstf = self.ap.read32(self.address + TraceFunnel.CSTF)
        cstf = cstf & ~TraceFunnel.CSTF_ENSX_MASK
        cstf = cstf | channels
        self.ap.write32(self.address + TraceFunnel.CSTF, cstf)

        return True

    def enable(self) -> None:
        """@brief Enables all channels"""
        self.set_enabled_channels(2**self.available_channels-1)

    def disable(self) -> None:
        """@brief Disables all channels"""
        self.set_enabled_channels(0x00)



# pyOCD debugger
# Copyright (c) 2018-2019 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from libusb_package import find as usb_find
except ImportError:
    from usb.core import find as usb_find
import usb.core
import usb.util
import logging
from typing import NamedTuple
import platform
import errno
from binascii import hexlify

from ...core import exceptions
from .. import common

LOG = logging.getLogger(__name__)

TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

class STLinkInfo(NamedTuple):
    """@brief STLink USB interface numbers and version name."""
    version_name: str
    out_ep: int
    in_ep: int
    swv_ep: int

class STLinkUSBInterface:
    """@brief Provides low-level USB enumeration and transfers for STLinkV2/3 devices."""

    ## Command packet size.
    CMD_SIZE = 16

    ## ST's USB vendor ID
    USB_VID = 0x0483

    ## Map of USB PID to firmware version name and device endpoints.
    #
    # Other PIDs:
    # - 0x3744: STLink V1
    # - 0x374d: STLink V3 DFU
    USB_PID_EP_MAP = {
        # PID              Version  OUT     IN      SWV
        0x3748: STLinkInfo('V2',    0x02,   0x81,   0x83),
        0x374a: STLinkInfo('V2-1',  0x01,   0x81,   0x82),  # Audio
        0x374b: STLinkInfo('V2-1',  0x01,   0x81,   0x82),
        0x374e: STLinkInfo('V3',    0x01,   0x81,   0x82),
        0x374f: STLinkInfo('V3',    0x01,   0x81,   0x82),  # Bridge
        0x3752: STLinkInfo('V2-1',  0x01,   0x81,   0x82),  # No MSD
        0x3753: STLinkInfo('V3',    0x01,   0x81,   0x82),  # 2VCP, No MSD
        0x3754: STLinkInfo('V3',    0x01,   0x81,   0x82),  # No MSD
        0x3755: STLinkInfo('V3',    0x01,   0x81,   0x82),
        0x3757: STLinkInfo('V3',    0x01,   0x81,   0x82),
        }

    ## STLink devices only have one USB interface.
    DEBUG_INTERFACE_NUMBER = 0

    @classmethod
    def _usb_match(cls, dev):
        try:
            # Check VID/PID.
            isSTLink = (dev.idVendor == cls.USB_VID) and (dev.idProduct in cls.USB_PID_EP_MAP)

            # Try accessing the current config, which will cause a permission error on Linux. Better
            # to error out here than later when building the device description. For Windows we
            # don't need to worry about device permissions, but reading descriptors requires special
            # handling due to the libusb bug described in __init__().
            if isSTLink and platform.system() != "Windows":
                dev.get_active_configuration()

            return isSTLink
        except usb.core.USBError as error:
            if error.errno == errno.EACCES and platform.system() == "Linux" \
                and common.should_show_libusb_device_error((dev.idVendor, dev.idProduct)):
                # We've already checked that this is an STLink device by VID/PID, so we
                # can use a warning log level to let the user know it's almost certainly
                # a permissions issue.
                LOG.warning("%s while trying to get the STLink USB device configuration "
                   "(VID=%04x PID=%04x). This can probably be remedied with a udev rule. "
                   "See <https://github.com/pyocd/pyOCD/tree/master/udev> for help.",
                   error, dev.idVendor, dev.idProduct)
            return False
        except (IndexError, NotImplementedError, ValueError) as error:
            return False

    @classmethod
    def get_all_connected_devices(cls):
        try:
            devices = usb_find(find_all=True, custom_match=cls._usb_match)
        except usb.core.NoBackendError:
            common.show_no_libusb_warning()
            return []

        assert devices is not None
        intfList = []
        for dev in devices:
            try:
                intf = cls(dev)
                intfList.append(intf)
            except (ValueError, usb.core.USBError, IndexError, NotImplementedError) as error:
                # Ignore errors that can be raised by libusb, just don't add the device to the list.
                pass

        return intfList

    def __init__(self, dev):
        self._dev = dev
        assert dev.idVendor == self.USB_VID
        self._info = self.USB_PID_EP_MAP[dev.idProduct]
        self._ep_out = None
        self._ep_in = None
        self._ep_swv = None
        self._max_packet_size = 64
        self._closed = True

        # Open the device temporarily to read the descriptor strings. The Windows libusb
        # (version 1.0.22 at the time of this writing) appears to have a bug where it can fail to
        # properly close a device automatically opened for reading descriptors. The bug manifests
        # as every other call to get_all_connected_devices() returning no available probes,
        # caused by a getting a permissions error ("The device has no langid" ValueError) when
        # attempting to read descriptor strings. If we manually call dispose_resources() after
        # reading the strings, everything is ok. This workaround doesn't cause any issues with
        # Linux or macOS.
        try:
            if len(self._dev.serial_number) == 12:  # Workaround for unprintable characters in the ST-Link V2 probes
                self._serial_number = hexlify(self._dev.serial_number.encode('utf-16-le')[::2])\
                    .decode('utf-8', 'replace').upper()
            else:
                self._serial_number = self._dev.serial_number
            self._vendor_name = self._dev.manufacturer
            self._product_name = self._dev.product
        finally:
            usb.util.dispose_resources(self._dev)

    def open(self):
        assert self._closed

        # Debug interface is always interface 0, alt setting 0.
        config = self._dev.get_active_configuration()
        interface = config[(self.DEBUG_INTERFACE_NUMBER, 0)]

        # Look up endpoint objects.
        for endpoint in interface:
            if endpoint.bEndpointAddress == self._info.out_ep:
                self._ep_out = endpoint
            elif endpoint.bEndpointAddress == self._info.in_ep:
                self._ep_in = endpoint
            elif endpoint.bEndpointAddress == self._info.swv_ep:
                self._ep_swv = endpoint

        if not self._ep_out:
            raise exceptions.ProbeError("Unable to find OUT endpoint")
        if not self._ep_in:
            raise exceptions.ProbeError("Unable to find IN endpoint")

        self._max_packet_size = self._ep_in.wMaxPacketSize

        # Claim this interface to prevent other processes from accessing it.
        usb.util.claim_interface(self._dev, self.DEBUG_INTERFACE_NUMBER)

        self._flush_rx()
        self._closed = False

    def close(self):
        assert not self._closed
        self._closed = True
        usb.util.release_interface(self._dev, self.DEBUG_INTERFACE_NUMBER)
        usb.util.dispose_resources(self._dev)
        self._ep_out = None
        self._ep_in = None

    @property
    def is_open(self) -> bool:
        return not self._closed

    @property
    def serial_number(self):
        return self._serial_number

    @property
    def vendor_name(self):
        return self._vendor_name

    @property
    def product_name(self):
        return self._product_name

    @property
    def version_name(self):
        return self._info.version_name

    @property
    def max_packet_size(self):
        return self._max_packet_size

    def _flush_rx(self):
        assert self._ep_in

        # Flush the RX buffers by reading until timeout exception
        try:
            while True:
                self._ep_in.read(self._max_packet_size, 1)
        except usb.core.USBError:
            # USB timeout expected
            pass

    def _read(self, size, timeout=1000):
        assert self._ep_in

        # Minimum read size is the maximum packet size.
        read_size = max(size, self._max_packet_size)
        data = self._ep_in.read(read_size, timeout)
        return bytearray(data)[:size]

    def transfer(self, cmd, writeData=None, readSize=None, timeout=1000):
        assert self._ep_out

        # Pad command to required 16 bytes.
        assert len(cmd) <= self.CMD_SIZE
        paddedCmd = bytearray(self.CMD_SIZE)
        paddedCmd[0:len(cmd)] = cmd

        try:
            # Command phase.
            if TRACE.isEnabledFor(logging.DEBUG):
                TRACE.debug("  USB CMD> (%d) %s", len(paddedCmd), ' '.join([f'{i:02x}' for i in paddedCmd]))
            count = self._ep_out.write(paddedCmd, timeout)
            assert count == len(paddedCmd)

            # Optional data out phase.
            if writeData is not None:
                if TRACE.isEnabledFor(logging.DEBUG):
                    TRACE.debug("  USB OUT> (%d) %s", len(writeData), ' '.join([f'{i:02x}' for i in writeData]))
                count = self._ep_out.write(writeData, timeout)
                assert count == len(writeData)

            # Optional data in phase.
            if readSize is not None:
                if TRACE.isEnabledFor(logging.DEBUG):
                    TRACE.debug("  USB IN < (req %d bytes)", readSize)
                data = self._read(readSize)
                if TRACE.isEnabledFor(logging.DEBUG):
                    TRACE.debug("  USB IN < (%d) %s", len(data), ' '.join([f'{i:02x}' for i in data]))

                # Verify we got all requested data.
                if len(data) < readSize:
                    raise exceptions.ProbeError("received incomplete command response from STLink "
                            f"(got {len(data)}, expected {readSize}")

                return data
        except usb.core.USBError as exc:
            raise exceptions.ProbeError("USB Error: %s" % exc) from exc
        return None

    def read_swv(self, size, timeout=1000):
        assert self._ep_swv
        return bytearray(self._ep_swv.read(size, timeout))

    def __repr__(self):
        return "<{} @ {:#x} vid={:#06x} pid={:#06x} sn={} version={}>".format(
            self.__class__.__name__, id(self),
            self._dev.idVendor, self._dev.idProduct, self.serial_number,
            self.version_name)



# pyOCD debugger
# Copyright (c) 2013-2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from time import sleep

from pyocd.coresight.generic_mem_ap import GenericMemAPTarget
from ...core import exceptions
from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (MemoryMap, RamRegion)
from ...core.target import Target
from ...coresight.cortex_m import CortexM
from ...utility.timeout import Timeout

LOG = logging.getLogger(__name__)


class CortexM_PSoC6(CortexM):
    VTBASE_CM0 = None
    VTBASE_CM4 = None

    def reset(self, reset_type=None):
        if reset_type is not Target.ResetType.HW:
            self.session.notify(Target.Event.PRE_RESET, self)
        self._run_token += 1
        if reset_type is Target.ResetType.HW:
            self._ap.dp.reset()
            sleep(0.5)
            self._ap.dp.connect()
            self.fpb.enable()
        else:
            if reset_type is Target.ResetType.SW_VECTRESET:
                mask = CortexM.NVIC_AIRCR_VECTRESET
            else:
                mask = CortexM.NVIC_AIRCR_SYSRESETREQ

            try:
                self.write_memory(CortexM.NVIC_AIRCR, CortexM.NVIC_AIRCR_VECTKEY | mask)
                self.flush()
            except exceptions.TransferError:
                self.flush()

        with Timeout(5.0) as t_o:
            while t_o.check():
                try:
                    dhcsr_reg = self.read32(CortexM.DHCSR)
                    if (dhcsr_reg & CortexM.S_RESET_ST) == 0:
                        break
                except exceptions.TransferError:
                    self.flush()
                    try:
                        self._ap.dp.connect()
                    except exceptions.TransferError:
                        self.flush()

                    sleep(0.01)

        if reset_type is not Target.ResetType.HW:
            self.session.notify(Target.Event.POST_RESET, self)

    def wait_halted(self):
        with Timeout(5.0) as t_o:
            while t_o.check():
                try:
                    if not self.is_running():
                        break
                except exceptions.TransferError:
                    self.flush()
                    sleep(0.01)
            else:
                raise exceptions.TimeoutError("Timeout waiting for target halt")

    def reset_and_halt(self, reset_type=None):
        self.halt()
        self.reset(reset_type)
        self.halt()
        self.wait_halted()

        if self.core_number == 0:
            vtbase = self.read_memory(self.VTBASE_CM0)
        elif self.core_number == 1:
            vtbase = self.read_memory(self.VTBASE_CM4)
        else:
            raise exceptions.TargetError("Invalid CORE ID")

        vtbase &= 0xFFFFFF00
        if vtbase < 0x10000000 or vtbase > 0x10200000:
            LOG.info("Vector Table address invalid (0x%08X), will not halt at main()", vtbase)
            return

        entry = self.read_memory(vtbase + 4)
        if entry < 0x10000000 or entry > 0x10200000:
            LOG.info("Entry Point address invalid (0x%08X), will not halt at main()", entry)
            return

        self.set_breakpoint(entry)
        self.bp_manager.flush()
        self.reset(self.ResetType.SW_SYSRESETREQ)
        sleep(0.2)
        self.wait_halted()
        self.remove_breakpoint(entry)
        self.bp_manager.flush()


class CortexM_PSoC6_BLE2(CortexM_PSoC6):
    VTBASE_CM0 = 0x402102B0
    VTBASE_CM4 = 0x402102C0


class CortexM_PSoC6_A2M(CortexM_PSoC6):
    VTBASE_CM0 = 0x40201120
    VTBASE_CM4 = 0x40200200


class PSoC6(CoreSightTarget):
    VENDOR = "Cypress"
    cortex_m_core_class = None

    def __init__(self, session, cortex_m_core_class, memory_map):
        self.cortex_m_core_class = cortex_m_core_class
        super(PSoC6, self).__init__(session, memory_map)

    def create_init_sequence(self):
        seq = super(PSoC6, self).create_init_sequence()
        seq.wrap_task('discovery',
            lambda seq: seq.replace_task('create_cores', self.create_psoc_cores)
            )
        return seq

    def create_psoc_cores(self):
        core0 = self.cortex_m_core_class(self.session, self.aps[1], self.memory_map, 0)
        core0.default_reset_type = self.ResetType.SW_SYSRESETREQ
        core1 = self.cortex_m_core_class(self.session, self.aps[2], self.memory_map, 1)
        core1.default_reset_type = self.ResetType.SW_SYSRESETREQ

        self.aps[1].core = core0
        self.aps[2].core = core1
        core0.init()
        core1.init()
        self.add_core(core0)
        self.add_core(core1)


class CortexM_PSoC64(CortexM):
    TEST_MODE_ADDR = None
    MAGIC_NUM_ADDR = None
    CM4_PWR_CTL_ADDR = None

    MAGIC_NUM_VALUE = 0x12344321
    TEST_MODE_VALUE = 0x80000000
    CM4_PWR_CTL_VALUE = 0x05FA0003

    def __init__(self, session, ap, memory_map, core_num, acquire_timeout):
        self._acquire_timeout = acquire_timeout
        self._skip_reset_and_halt = False
        super(CortexM_PSoC64, self).__init__(session, ap, memory_map, core_num)

    @property
    def acquire_timeout(self):
        return self._acquire_timeout

    @acquire_timeout.setter
    def acquire_timeout(self, value):
        self._acquire_timeout = value

    @property
    def skip_reset_and_halt(self):
        return self._skip_reset_and_halt

    @skip_reset_and_halt.setter
    def skip_reset_and_halt(self, value):
        self._skip_reset_and_halt = value

    def reset(self, reset_type=None):
        if reset_type is not Target.ResetType.HW:
            self.session.notify(Target.Event.PRE_RESET, self)

        self._run_token += 1

        if reset_type is Target.ResetType.HW:
            self._ap.dp.reset()
            self.reinit_dap()
            self.fpb.enable()

        else:
            if reset_type is Target.ResetType.SW_VECTRESET:
                mask = CortexM.NVIC_AIRCR_VECTRESET
            else:
                mask = CortexM.NVIC_AIRCR_SYSRESETREQ

            try:
                self.write_memory(CortexM.NVIC_AIRCR, CortexM.NVIC_AIRCR_VECTKEY | mask)
                self.flush()
            except exceptions.TransferError:
                self.flush()

        with Timeout(5.0) as t_o:
            while t_o.check():
                try:
                    self._ap.dp.connect()
                    dhcsr_reg = self.read32(CortexM.DHCSR)
                    if (dhcsr_reg & CortexM.S_RESET_ST) == 0:
                        break
                    self.flush()
                except exceptions.TransferError:
                    pass

        if reset_type is not Target.ResetType.HW:
            self.session.notify(Target.Event.POST_RESET, self)

    def wait_halted(self):
        with Timeout(5.0) as t_o:
            while t_o.check():
                try:
                    if not self.is_running():
                        break
                except exceptions.TransferError:
                    self.flush()
                    sleep(0.01)
            else:
                raise Exception("Timeout waiting for target halt")

    def reinit_dap(self):
        with Timeout(2.0) as t_o:
            while t_o.check():
                try:
                    self._ap.dp.connect()
                    self.flush()
                    break
                except exceptions.TransferError:
                    self.flush()

            else:
                LOG.error("Failed to initialize DAP")

    def acquire(self):
        with Timeout(self.acquire_timeout) as t_o:
            while t_o.check():
                try:
                    self._ap.dp.connect()
                    # self.write32(self.IPC2_DATA_ADDR, 0)
                    self.write32(self.TEST_MODE_ADDR, self.TEST_MODE_VALUE)
                    self.flush()
                    break
                except exceptions.TransferError:
                    pass

            else:
                LOG.warning("Failed to enter test mode")

    def check_flashboot_ver(self):
        fb_ver_hi = self.read32(0x16002004)
        fb_ver_lo = self.read32(0x16002018)
        b0 = fb_ver_hi >> 28
        b1 = (fb_ver_hi >> 24) & 0xF
        b2 = (fb_ver_hi >> 16) & 0xFF
        b3 = fb_ver_hi & 0x0000FFFF
        if b3 != 0x8001 or b0 != 2:
            LOG.error("Flash Boot is corrupted or non Flash Boot image programmed")
            return

        patch = fb_ver_lo >> 24
        if b1 == 4 and b2 == 0 and patch == 0:
            LOG.warning("Pre-production version of device is detected which is incompatible with this software")
            LOG.warning("Please contact Cypress for new production parts")

        return

    def reset_and_halt(self, reset_type=None):
        if self.skip_reset_and_halt:
            return

        LOG.info("Acquiring target...")

        self.reset(self.ResetType.SW_SYSRESETREQ)
        try:
            self.flush()
        except exceptions.TransferError:
            pass

        self.acquire()

        with Timeout(self.acquire_timeout) as t_o:
            while t_o.check():
                try:
                    if self.read32(self.MAGIC_NUM_ADDR) == self.MAGIC_NUM_VALUE:
                        break
                except exceptions.TransferError:
                    pass

        if not t_o.check():
            LOG.warning("Failed to acquire the target (listen window not implemented?)")

        self.check_flashboot_ver()

        with Timeout(self.acquire_timeout) as t_o:
            while t_o.check():
                try:
                    self._ap.dp.connect()
                    self.halt()
                    self.wait_halted()
                    self.write_core_register('xpsr', CortexM.XPSR_THUMB)
                    break
                except exceptions.TransferError:
                    pass

    def resume(self):
        from .flash_psoc6 import Flash_PSoC64
        super(CortexM_PSoC64, self).resume()

        if not Flash_PSoC64.isFlashing:
            LOG.info("Clearing TEST_MODE bit...")
            self.write32(self.TEST_MODE_ADDR, 0)
            self.flush()


class CortexM_PSoC64_BLE2(CortexM_PSoC64):
    TEST_MODE_ADDR = 0x40260100
    MAGIC_NUM_ADDR = 0x08044804
    CM4_PWR_CTL_ADDR = 0x40210080


class CortexM_PSoC64_A2M(CortexM_PSoC64):
    TEST_MODE_ADDR = 0x40260100
    MAGIC_NUM_ADDR = 0x080FE004
    CM4_PWR_CTL_ADDR = 0x40201200


class CortexM_PSoC64_A512K(CortexM_PSoC64):
    TEST_MODE_ADDR = 0x40260100
    MAGIC_NUM_ADDR = 0x0803E004
    CM4_PWR_CTL_ADDR = 0x40201200


class SYS_AP_PSoC64(GenericMemAPTarget):
    def __init__(self, session, ap, memory_map, core_num, acquire_timeout):
        self._acquire_timeout = acquire_timeout
        self._skip_reset_and_halt = False
        super(SYS_AP_PSoC64, self).__init__(session, ap, memory_map, core_num)

    @property
    def acquire_timeout(self):
        return self._acquire_timeout

    @acquire_timeout.setter
    def acquire_timeout(self, value):
        self._acquire_timeout = value

    @property
    def skip_reset_and_halt(self):
        return self._skip_reset_and_halt

    @skip_reset_and_halt.setter
    def skip_reset_and_halt(self, value):
        self._skip_reset_and_halt = value

    def reset(self, reset_type=None):
        self.session.notify(Target.Event.PRE_RESET, self)
        try:
            self.ap.dp.reset()
        except exceptions.TransferError:
            pass

        with Timeout(self._acquire_timeout) as t_o:
            while t_o.check():
                try:
                    self._ap.dp.connect()
                    break
                except exceptions.TransferError:
                    pass

        self.session.notify(Target.Event.POST_RESET, self)


class PSoC64(CoreSightTarget):
    VENDOR = "Cypress"
    AP_NUM = None
    cortex_m_core_class = None

    def __init__(self, session, cortex_m_core_class, memory_map, ap_num):
        self.cortex_m_core_class = cortex_m_core_class
        self.AP_NUM = ap_num
        self.DEFAULT_ACQUIRE_TIMEOUT = 25.0
        super(PSoC64, self).__init__(session, memory_map)

    def create_init_sequence(self):
        seq = super(PSoC64, self).create_init_sequence()
        seq.wrap_task('discovery',
            lambda seq: seq
                        .replace_task('find_aps', self.find_aps)
                        .replace_task('create_cores', self.create_psoc_core)
            )
        return seq

    def find_aps(self):
        if self.dp.valid_aps is not None:
            return

        if self.AP_NUM:
            self.dp.valid_aps = (0, self.AP_NUM,)
        else:
            self.dp.valid_aps = (0,)

    def create_psoc_core(self):
        sysap = SYS_AP_PSoC64(self.session, self.aps[0], self.memory_map, 0, self.DEFAULT_ACQUIRE_TIMEOUT)
        sysap.default_reset_type = self.ResetType.SW_SYSRESETREQ
        self.aps[0].core = sysap
        sysap.init()
        self.add_core(sysap)

        if self.AP_NUM:
            core = self.cortex_m_core_class(self.session, self.aps[self.AP_NUM], self.memory_map, 1,
                                     self.DEFAULT_ACQUIRE_TIMEOUT)
            core.default_reset_type = self.ResetType.SW_SYSRESETREQ
            self.aps[self.AP_NUM].core = core
            core.init()
            self.add_core(core)
            self.selected_core = 1


class cy8c64_sysap(PSoC64):
    MEMORY_MAP = MemoryMap(RamRegion(start=0, length=0x100000000))

    def __init__(self, session):
        super(cy8c64_sysap, self).__init__(session, None, self.MEMORY_MAP, 0)



# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import (Any, Callable, Dict, Iterable, List, Optional, Tuple, Union, cast)

from ..core.target import Target
from ..core.options import OPTIONS_INFO
from ..utility.compatibility import to_str_safe

LOG = logging.getLogger(__name__)

def split_command(cmd: str) -> List[str]:
    """@brief Split command by whitespace, supporting quoted strings."""
    result: List[str] = []
    state = 0
    word = ''
    open_quote = ''
    pos = 0
    while pos < len(cmd):
        c = cmd[pos]
        pos += 1
        if state == 0:
            if c.isspace():
                if word:
                    result.append(word)
                    word = ''
            elif c in ('"', "'"):
                if word:
                    result.append(word)
                word = ''
                open_quote = c
                state = 1
            elif c in ';!@#$%^&*()+=[]{}|<>,?':
                if word:
                    result.append(word)
                word = c
                state = 2
            elif c == '\\':
                if pos < len(cmd):
                    c = cmd[pos]
                    pos += 1
                    word += c
            else:
                word += c
        elif state == 1:
            if c == open_quote:
                result.append(word)
                word = ''
                state = 0
            # Only honour escapes in double quotes.
            elif open_quote == '"' and c == '\\':
                if pos < len(cmd):
                    c = cmd[pos]
                    pos += 1
                    word += c
            else:
                word += c
        elif state == 2:
            if word:
                result.append(word)
            # Back up to reprocess this char in state 0.
            word = ''
            pos -= 1
            state = 0
    if word:
        result.append(word)
    return result

def split_command_line(cmd_line: Union[str, List[str]]) -> List[str]:
    """@brief Split command line by whitespace, supporting quoted strings."""
    result: List[str] = []
    if isinstance(cmd_line, str):
        args = [cmd_line]
    else:
        args = cmd_line
    for cmd in args:
        result += split_command(cmd)
    return result

## Map of vector char characters to masks.
VECTOR_CATCH_CHAR_MAP = {
        'e': Target.VectorCatch.SECURE_FAULT,
        'h': Target.VectorCatch.HARD_FAULT,
        'b': Target.VectorCatch.BUS_FAULT,
        'm': Target.VectorCatch.MEM_FAULT,
        'i': Target.VectorCatch.INTERRUPT_ERR,
        's': Target.VectorCatch.STATE_ERR,
        'c': Target.VectorCatch.CHECK_ERR,
        'p': Target.VectorCatch.COPROCESSOR_ERR,
        'r': Target.VectorCatch.CORE_RESET,
        'a': Target.VectorCatch.ALL,
        'n': Target.VectorCatch.NONE,
    }

def convert_vector_catch(vcvalue: Union[str, bytes]) -> int:
    """@brief Convert a vector catch string to a mask.

    @exception ValueError Raised if an invalid vector catch character is encountered.
    """
    # Make case insensitive.
    value: str = to_str_safe(vcvalue).lower()

    # Handle special vector catch options.
    if value == 'all':
        return Target.VectorCatch.ALL
    elif value == 'none':
        return Target.VectorCatch.NONE

    # Convert options string to mask.
    try:
        return sum([VECTOR_CATCH_CHAR_MAP[c] for c in value])
    except KeyError as e:
        # Reraise an error with a more helpful message.
        raise ValueError("invalid vector catch option '{}'".format(e.args[0]))

def _convert_string_list_option(value: Optional[str]) -> List[str]:
    """@brief Convert a comma-separated list of strings.

    A None value results in an empty list.

    No special support for quoted values is provided, so a comma in between quotes is treated
    the same as a comma elsewhere.
    """
    if value is None:
        return []
    return [
        i.strip()
        for i in value.split(',')
    ]

## Map with special converter routines for session options that need them.
_OPTION_CONVERTERS: Dict[str, Callable[[Optional[str]], Any]] = {
    'pack.debug_sequences.disabled_sequences': _convert_string_list_option,
}

def convert_one_session_option(name: str, value: Optional[str]) -> Tuple[str, Any]:
    """@brief Convert one session option's value from a string.

    Handles "no-" prefixed option names by inverting their boolean value. If a non-boolean option has
    a "no-" prefix, then a warning is logged and None returned for the value.

    @return Bi-tuple of option name, converted value. The option name may be modified from the one passed
        in for cases like a "no-" prefix.
    """
    # Check for and strip "no-" prefix before we validate the option name.
    if name.startswith('no-'):
        name = name[3:]
        had_no_prefix = True
    else:
        had_no_prefix = False

    # Look up this option.
    try:
        info = OPTIONS_INFO[name]
    except KeyError:
        # Return the value unmodified for unknown options.
        LOG.warning("unknown session option '%s'", name)
        return name, value

    # Check if "no-" prefix is allowed. Only bool options can use it.
    if had_no_prefix and (info.type is not bool):
        LOG.warning("'no-' prefix used on non-boolean session option '%s'", name)
        return name, None

    # Check for a special converter function.
    if name in _OPTION_CONVERTERS:
        return name, _OPTION_CONVERTERS[name](value)

    # Default result; unset option value.
    result = None

    # Extract the option's type. If its type is a tuple of types, then take the first type.
    if isinstance(info.type, tuple):
        option_type = cast(tuple, info.type)[0]
    else:
        option_type = info.type

    # Handle bool options without a value specially.
    if value is None:
        if issubclass(option_type, bool):
            result = not had_no_prefix
        else:
            LOG.warning("non-boolean option '%s' requires a value", name)
    # Convert string value to option type.
    elif issubclass(option_type, bool):
        if value.lower() in ("true", "1", "yes", "on", "false", "0", "no", "off"):
            result = value.lower() in ("true", "1", "yes", "on")

            # If a bool option with "no-" prefix has a value, the value is inverted.
            if had_no_prefix:
                result = not result
        else:
            LOG.warning("invalid value for option '%s'", name)
    elif issubclass(option_type, int):
        try:
            result = int(value, base=0)
        except ValueError:
            LOG.warning("invalid value for option '%s'", name)
    elif issubclass(option_type, float):
        try:
            result = float(value)
        except ValueError:
            LOG.warning("invalid value for option '%s'", name)
    else:
        result = value

    return name, result

def convert_session_options(option_list: Iterable[str]) -> Dict[str, Any]:
    """@brief Convert a list of session option settings to a dictionary."""
    options = {}
    if option_list is not None:
        for o in option_list:
            if '=' in o:
                name, value = o.split('=', 1)
                name = name.strip().lower()
                value = value.strip()
            else:
                name = o.strip().lower()
                value = None

            name, value = convert_one_session_option(name, value)
            if value is not None:
                options[name] = value
    return options

## Map to convert from reset type names to enums.
RESET_TYPE_MAP: Dict[str, Optional[Target.ResetType]] = {
        'default': None,
        'hw': Target.ResetType.HW,
        'sw': Target.ResetType.SW,
        'hardware': Target.ResetType.HW,
        'software': Target.ResetType.SW,
        'sw_system': Target.ResetType.SW_SYSTEM,
        'sw_core': Target.ResetType.SW_CORE,
        'sw_sysresetreq': Target.ResetType.SW_SYSRESETREQ,
        'sw_vectreset': Target.ResetType.SW_VECTRESET,
        'sw_emulated': Target.ResetType.SW_EMULATED,
        'system': Target.ResetType.SW_SYSTEM,
        'core': Target.ResetType.SW_CORE,
        'sysresetreq': Target.ResetType.SW_SYSRESETREQ,
        'vectreset': Target.ResetType.SW_VECTRESET,
        'emulated': Target.ResetType.SW_EMULATED,
    }

def convert_reset_type(value: str) -> Optional[Target.ResetType]:
    """@brief Convert a reset_type session option value to the Target.ResetType enum.
    @param value The value of the reset_type session option.
    @exception ValueError Raised if an unknown reset_type value is passed.
    """
    value = value.lower()
    if value not in RESET_TYPE_MAP:
        raise ValueError("unexpected value for reset_type option ('%s')" % value)
    return RESET_TYPE_MAP[value]

def convert_frequency(value: str) -> int:
    """@brief Applies scale suffix to frequency value string.
    @param value String with a float and possible 'k' or 'm' suffix (case-insensitive). "Hz" may
        also follow. No space is allowed between the float and suffix. Leading and trailing
        whitespace is allowed.
    @return Integer scaled according to optional metric suffix.
    """
    value = value.strip().lower()
    if value.endswith("hz"):
        value = value[:-2]
    suffix = value[-1]
    if suffix in ('k', 'm'):
        fvalue = float(value[:-1])
        if suffix == 'k':
            fvalue *= 1000
        elif suffix == 'm':
            fvalue *= 1000000
        return int(fvalue)
    else:
        return int(float(value))


def int_base_0(x: str) -> int:
    """@brief Converts a string to an int with support for base prefixes."""
    return int(x, base=0)


def flatten_args(args: Iterable[Iterable[Any]]) -> List[Any]:
    """@brief Converts a list of lists to a single list."""
    return [item for sublist in args for item in sublist]




# pyOCD debugger
# Copyright (c) 2006-2013 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...debug.svd.loader import SVDFile

FLASH_ALGO = { 'load_address' : 0x20000000,
               'instructions' : [
    0xE00ABE00, 0x062D780D, 0x24084068, 0xD3000040, 0x1E644058, 0x1C49D1FA, 0x2A001E52, 0x4770D1F2,
    0x21002210, 0xf8c00690, 0x4a2c2630, 0xf0236813, 0x60134380, 0x1280f8c0, 0x1284f8c0, 0x68134a28,
    0x4370f423, 0xf8c06013, 0x21021380, 0x20006001, 0x20004770, 0xb5104770, 0x2000210f, 0xf844f000,
    0xbf182800, 0x4a1fbd10, 0xe8bd210f, 0x20004010, 0xb864f000, 0x0bc4b510, 0x46084621, 0xf834f000,
    0xbf182800, 0x4a17bd10, 0xe8bd4621, 0x46084010, 0xb854f000, 0x4614b570, 0xd10e0005, 0x0100e9d4,
    0xe9d44408, 0x44111202, 0x69214408, 0x69614408, 0x69a14408, 0x42404408, 0x0be861e0, 0xf0004601,
    0x2800f813, 0xbd70bf18, 0x46214b06, 0xe8bd4628, 0xf44f4070, 0xf0007280, 0x0000b818, 0x40000500,
    0x40000400, 0x00b71b00, 0xb08bb500, 0x92002232, 0x0101e9cd, 0x46684a39, 0x4790a906, 0x28009806,
    0xf600bf18, 0xb00b10c4, 0xb500bd00, 0xf04fb08b, 0x92030c33, 0xc000f8cd, 0x0101e9cd, 0x707af44f,
    0xf0f0fbb3, 0x4a2d9004, 0xa9064668, 0x98064790, 0xbf182800, 0x10c4f600, 0xbd00b00b, 0xb08bb500,
    0x93002334, 0x0101e9cd, 0x707af44f, 0xf0f0fbb2, 0x4a229003, 0xa9064668, 0x98064790, 0xbf182800,
    0x10c4f600, 0xbd00b00b, 0xb08bb500, 0x9300233b, 0x0101e9cd, 0x707af44f, 0xf0f0fbb2, 0x4a179003,
    0xa9064668, 0x98064790, 0xbf182800, 0x10c4f600, 0xbd00b00b, 0xb08bb500, 0x92002235, 0x0101e9cd,
    0x46684a0e, 0x4790a906, 0x28009806, 0xf600bf18, 0xb00b10c4, 0xb500bd00, 0x2338b08b, 0x93009203,
    0x0101e9cd, 0x46684a05, 0x4790a906, 0x28009806, 0xf600bf18, 0xb00b10c4, 0x0000bd00, 0x03000205,
    0x00000000
                                ],

    # Relative function addresses
    'pc_init': 0x20000021,
    'pc_program_page': 0x20000095,
    'pc_erase_sector': 0x20000075,
    'pc_eraseAll': 0x20000057,
    'begin_stack' : 0x20001000,
    'begin_data' : 0x20003000,
    'page_buffers' : [0x20003000],
    'static_base' : 0x200001dc,
    'page_size' : 0x00000100,
    'min_program_length' : 256,
    'analyzer_supported' : False
}

class LPC54608(CoreSightTarget):

    VENDOR = "NXP"

    MEMORY_MAP = MemoryMap(
        FlashRegion(name='flash',   start=0,           length=0x80000,  is_boot_memory=True,
                                                                        blocksize=0x8000,
                                                                        page_size=0x100,
                                                                        algo=FLASH_ALGO),
        RamRegion(  name='sramx',   start=0x04000000,  length=0x8000),
        RamRegion(  name='sram0',   start=0x20000000,  length=0x10000),
        RamRegion(  name='sram1',   start=0x20010000,  length=0x10000),
        RamRegion(  name='sram2',   start=0x20020000,  length=0x8000)
        )

    def __init__(self, session):
        super(LPC54608, self).__init__(session, self.MEMORY_MAP)
        self.ignoreReset = False
        self._svd_location = SVDFile.from_builtin("LPC54608.xml")

    def reset_and_halt(self, reset_type=None, map_to_user=True):
        super(LPC54608, self).reset_and_halt(reset_type)

        # Remap to use flash and set SP and SP accordingly
        if map_to_user:
            self.write_memory(0x40000000, 0x2, 32)
            sp = self.read_memory(0x0)
            pc = self.read_memory(0x4)
            self.write_core_register('sp', sp)
            self.write_core_register('pc', pc)



# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import sys
from typing import (Any, Callable, cast, Dict, IO, Iterator, List, NamedTuple, Optional, Sequence,
        TYPE_CHECKING)
import six
import pprint
import subprocess
from shutil import get_terminal_size

from ..core import exceptions
from ..coresight.ap import MEM_AP
from ..utility.strings import UniquePrefixMatcher
from ..utility.cmdline import split_command_line

if TYPE_CHECKING:
    from ..debug.svd.model import SVDPeripheral
    from ..core.session import Session
    from ..core.core_target import CoreTarget
    from ..core.soc_target import SoCTarget
    from ..board.board import Board
    from ..coresight.ap import (APAddressBase, AccessPort)
    from ..coresight.coresight_target import CoreSightTarget
    from ..probe.debug_probe import DebugProbe

LOG = logging.getLogger(__name__)

class CommandSet:
    """@brief Holds a set of command classes."""

    ## Whether command and infos modules have been loaded yet.
    DID_LOAD_COMMAND_MODULES = False

    def __init__(self):
        self._commands = {} # Dict of all available commands.
        self._command_classes = set()
        self._command_matcher = UniquePrefixMatcher()
        self._values = {}
        self._value_classes = set()
        self._value_matcher = UniquePrefixMatcher()

        # Ensure these modules get loaded so the ALL_COMMANDS dicts are filled.
        self._load_modules()

    @classmethod
    def _load_modules(cls):
        # Load these modules in order to populate the dicts with standard commands. This must be
        # done lazily because the commands import lots of modules from pyocd that would cause import
        # cycles otherwise.
        if not cls.DID_LOAD_COMMAND_MODULES:
            from . import commands
            from . import values
            cls.DID_LOAD_COMMAND_MODULES = True

    @property
    def commands(self):
        return self._commands

    @property
    def command_classes(self):
        return self._command_classes

    @property
    def command_matcher(self):
        return self._command_matcher

    @property
    def values(self):
        return self._values

    @property
    def value_classes(self):
        return self._value_classes

    @property
    def value_matcher(self):
        return self._value_matcher

    def add_command_group(self, group_name):
        """@brief Add commands belonging to a group to the command set.
        @param self The command set.
        @param group_name String with the name of the group to add.
        """
        from .base import ALL_COMMANDS
        self.add_commands(ALL_COMMANDS.get(group_name, set()))

    def add_commands(self, commands):
        """@brief Add some commands to the command set.
        @param self The command set.
        @param commands List of command classes.
        """
        from .base import ValueBase
        value_classes = {klass for klass in commands if issubclass(klass, ValueBase)}
        cmd_classes = commands - value_classes
        cmd_names = {name: klass for klass in cmd_classes for name in klass.INFO['names']}
        self._commands.update(cmd_names)
        self._command_classes.update(cmd_classes)
        self._command_matcher.add_items(cmd_names.keys())

        value_names = {name: klass for klass in value_classes for name in klass.INFO['names']}
        self._values.update(value_names)
        self._value_classes.update(value_classes)
        self._value_matcher.add_items(value_names.keys())

class CommandInvocation(NamedTuple):
    """@brief Groups the command name with an iterable of args and a handler function.

    The handler is a callable that will evaluate the command. It accepts a single argument of the
    CommandInvocation instance.
    """
    cmd: str
    args: Sequence[str]
    handler: Callable[["CommandInvocation"], None] # type:ignore # mypy doesn't support recursive types yet!

class CommandExecutionContext:
    """@brief Manages command execution.

    This class holds persistent state for command execution, and provides the interface for executing
    commands and command lines.
    """

    _session: Optional[Session]
    _selected_core: Optional[CoreTarget]
    _selected_ap_address: Optional[APAddressBase]
    _peripherals: Dict[str, SVDPeripheral]
    _python_namespace: Dict[str, Any]

    def __init__(self, no_init: bool = False, output_stream: Optional[IO[str]] = None):
        """@brief Constructor.
        @param self This object.
        @param no_init Whether the board and target will be initialized when attach_session() is called.
            Defaults to False.
        @param output_stream Stream object to which command output and errors will be written. If not provided,
            output will be written to sys.stdout.
        @param elf_path Optional path to an ELF file.
        """
        self._no_init = no_init
        self._output = output_stream or sys.stdout
        self._python_namespace = {}
        self._command_set = CommandSet()

        # State attributes.
        self._session = None
        self._selected_core = None
        self._selected_ap_address = None
        self._peripherals = {}
        self._loaded_peripherals = False

        # Add in the standard commands.
        self._command_set.add_command_group('standard')

    def write(self, message='', **kwargs):
        """@brief Write a fixed message to the output stream.

        The message is written to the output stream passed to the constructor, terminated with
        a newline by default. The `end` keyword argument can be passed to change the terminator. No
        formatting is applied to the message. If formatting is required, use the writei() or writef()
        methods instead.

        @param self This object.
        @param message The text to write to the output. If not a string object, it is run through str().
        """
        if self._output is None:
            return
        end = kwargs.pop('end', "\n")
        if not isinstance(message, str):
            message = str(message)
        self._output.write(message + end)

    def writei(self, fmt, *args, **kwargs):
        """@brief Write an interpolated string to the output stream.

        The formatted string is written to the output stream passed to the constructor, terminated with
        a newline by default. The `end` keyword argument can be passed to change the terminator.

        @param self This object.
        @param fmt Format string using printf-style "%" formatters.
        """
        assert isinstance(fmt, str)
        message = fmt % args
        self.write(message, **kwargs)

    def writef(self, fmt, *args, **kwargs):
        """@brief Write a formatted string to the output stream.

        The formatted string is written to the output stream passed to the constructor, terminated with
        a newline by default. The `end` keyword argument can be passed to change the terminator.

        @param self This object.
        @param fmt Format string using the format() mini-language.
        """
        assert isinstance(fmt, str)
        message = fmt.format(*args, **kwargs)
        self.write(message, **kwargs)

    def attach_session(self, session):
        """@brief Associate a session with the command context.

        Various data for the context are initialized. This includes selecting the initially selected core and MEM-AP,
        and getting an ELF file that was set on the target.

        @param self This object.
        @param session A @ref pyocd.core.session.Session "Session" instance.
        @retval True Session attached and context state inited successfully.
        @retval False An error occurred when opening the session or initing the context state.
        """
        assert self._session is None
        assert session.is_open or self._no_init
        self._session = session
        assert self.target

        # Select the first core's MEM-AP by default.
        if not self._no_init:
            self.set_context_defaults()

        # Allow the target to add its own commands.
        self.target.add_target_command_groups(self.command_set)

        # Add user-defined commands once we know we have a session created.
        self.command_set.add_command_group('user')

        return True

    def set_context_defaults(self) -> None:
        """@brief Sets context attributes to their default values.

        Sets the selected core and selected MEM-AP to the default values.
        """
        assert self.target

        try:
            # Selected core defaults to the target's default selected core.
            if (self.selected_core is None) and (self.target.selected_core):
                self.selected_core = self.target.selected_core

            # Get the AP for the selected core.
            if self.selected_core is not None:
                self.selected_ap_address = self.selected_core.ap.address
        except IndexError:
            pass

        # Fall back to the first MEM-AP.
        if self.selected_ap_address is None:
            for ap_num in sorted(self.target.aps.keys()):
                if isinstance(self.target.aps[ap_num], MEM_AP):
                    self.selected_ap_address = ap_num
                    break

    @property
    def session(self) -> Session:
        assert self._session
        return self._session

    @property
    def board(self) -> Board:
        assert self._session and self._session.board
        return self._session.board

    @property
    def target(self) -> SoCTarget:
        assert self._session and self._session.target
        return self._session.target

    @property
    def probe(self) -> DebugProbe:
        assert self._session and self._session.probe
        return self._session.probe

    @property
    def elf(self):
        return self.target and self.target.elf

    @property
    def command_set(self):
        """@brief CommandSet with commands available in this context."""
        return self._command_set

    @property
    def peripherals(self):
        """@brief Dict of SVD peripherals."""
        assert self.target
        if self.target.svd_device and not self._loaded_peripherals:
            for p in self.target.svd_device.peripherals:
                self._peripherals[p.name.lower()] = p
            self._loaded_peripherals = True
        return self._peripherals

    @property
    def output_stream(self) -> IO[str]:
        return self._output

    @output_stream.setter
    def output_stream(self, stream: IO[str]) -> None:
        self._output = stream

    @property
    def selected_core(self) -> Optional[CoreTarget]:
        """@brief The Target instance for the selected core."""
        return self._selected_core

    @selected_core.setter
    def selected_core(self, value: CoreTarget) -> None:
        self._selected_core = value

    @property
    def selected_ap_address(self) -> Optional[APAddressBase]:
        return self._selected_ap_address

    @selected_ap_address.setter
    def selected_ap_address(self, value: APAddressBase) -> None:
        self._selected_ap_address = value

    @property
    def selected_ap(self) -> Optional[AccessPort]:
        if self.selected_ap_address is None:
            return None
        else:
            from ..coresight.coresight_target import CoreSightTarget
            assert self.target
            if isinstance(self.target, CoreSightTarget):
                return cast(CoreSightTarget, self.target).aps[self.selected_ap_address]
            else:
                raise exceptions.CommandError("target is not CoreSight based")

    def process_command_line(self, line: str) -> None:
        """@brief Run a command line consisting of one or more semicolon-separated commands.

        @param self
        @param line Complete command line string.
        """
        for args in self._split_commands(line):
            assert args
            invoc = self.parse_command(args)
            invoc.handler(invoc)

    def process_command_file(self, cmd_file: IO[str]) -> None:
        """@brief Run commands contained in a file.

        @param self
        @param cmd_file File object containing commands to run. Must be opened in text mode. When this method returns,
            the file will be closed. This is true even if an exception is raised during command execution.
        """
        try:
            for line in cmd_file:
                line = line.strip()

                # Skip empty or comment lines.
                if (len(line) == 0) or (line[0] == '#'):
                    continue

                self.process_command_line(line)
        finally:
            cmd_file.close()

    def _split_commands(self, line: str) -> Iterator[List[str]]:
        """@brief Generator yielding commands separated by semicolons.

        Python and system commands are handled specially. For these we yield a list of 2 elements: the command,
        either "$" or "!", followed by the unmodified remainder of the command line. For these commands,
        splitting on semicolons is not supported.
        """
        parts = split_command_line(line.strip())

        # Check for Python or system command. For these we yield a list of 2 elements: the command
        # followed by the rest of the command line as it was originally.
        if parts and (parts[0] in '$!'):
            # Remove the Python/system command prefix from the command line. Can't use str.removeprefix()
            # since it was added in 3.9.
            line_remainder = line.strip()
            assert line_remainder.find(parts[0]) == 0
            line_remainder = line_remainder[len(parts[0]):].strip()
            yield [parts[0], line_remainder]
            return

        result: List[str] = []

        for p in parts:
            if p == ';':
                if result:
                    yield result
                    result = []
            else:
                result.append(p)
        if result:
            yield result

    def parse_command(self, cmdline: List[str]) -> CommandInvocation:
        """@brief Create a CommandInvocation from a single command."""
        # Check for Python or system command lines.
        first_char = cmdline[0]
        if first_char in '$!':
            # cmdline parameters that are for Python and system commands must be a 2-element list,
            # as generated by _split_commands().
            assert len(cmdline) == 2

            # Return the invocation instance with the handler set appropriately.
            if first_char == '$':
                return CommandInvocation(cmdline[1], [], self.handle_python)
            elif first_char == '!':
                return CommandInvocation(cmdline[1], [], self.handle_system)

        # Split command into words.
        args = split_command_line(cmdline)
        cmd = args[0].lower()
        args = args[1:]

        # Look up shortened unambiguous match for the command name.
        matched_command = self._command_set.command_matcher.find_one(cmd)

        # Check for valid command.
        if matched_command is None:
            all_matches = self._command_set.command_matcher.find_all(cmd)
            if len(all_matches) > 1:
                raise exceptions.CommandError("command '%s' is ambiguous; matches are %s" % (cmd,
                        ", ".join("'%s'" % c for c in all_matches)))
            else:
                raise exceptions.CommandError("unrecognized command '%s'" % cmd)

        return CommandInvocation(matched_command, args, self.execute_command)

    def execute_command(self, invocation: CommandInvocation) -> None:
        """@brief Execute a single command."""
        # Must have an attached session to run commands, except for certain commands.
        assert (self.session is not None) or (invocation.cmd in ('list', 'help', 'exit'))

        # Run command.
        cmd_class = self._command_set.commands[invocation.cmd]
        cmd_object = cmd_class(self)
        cmd_object.check_arg_count(invocation.args)
        cmd_object.parse(invocation.args)

        if self.session:
            # Reroute print() in user-defined functions so it will come out our output stream.
            with self.session.user_script_print_proxy.push_target(self.write):
                cmd_object.execute()
        else:
            cmd_object.execute()

    def _build_python_namespace(self) -> None:
        """@brief Construct the dictionary used as the namespace for python commands."""
        assert self.session
        assert self.target
        ns = self.session.user_script_proxy.namespace
        ns.update({
                'elf': self.elf,
                'map': self.target.memory_map,
            })
        self._python_namespace = ns

    def handle_python(self, invocation: CommandInvocation) -> None:
        """@brief Evaluate a python expression."""
        assert self.session
        try:
            # Lazily build the python environment.
            if not self._python_namespace:
                self._build_python_namespace()

            # Reroute print() in user-defined functions so it will come out our output stream. Not that
            # we expect much use of print() from expressions...
            with self.session.user_script_print_proxy.push_target(self.write):
                result = eval(invocation.cmd, self._python_namespace)
                if result is not None:
                    if isinstance(result, int):
                        self.writei("0x%08x (%d)", result, result)
                    else:
                        w, h = get_terminal_size()
                        self.write(pprint.pformat(result, indent=2, width=w, depth=10))
        except Exception as e:
            # Log the traceback before raising the exception.
            if self.session and self.session.log_tracebacks:
                LOG.error("Exception while executing expression: %s", e, exc_info=True)
            raise exceptions.CommandError("exception while executing expression: %s" % e)

    def handle_system(self, invocation: CommandInvocation) -> None:
        """@brief Evaluate a system call command."""
        try:
            output = subprocess.check_output(invocation.cmd, stderr=subprocess.STDOUT, shell=True)
            self.write(six.ensure_str(output), end='')
        except subprocess.CalledProcessError as err:
            raise exceptions.CommandError(str(err)) from err



# pyOCD debugger
# Copyright (c) 2017-2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import logging

LOG = logging.getLogger(__name__)

class ProgressReport(object):
    """@brief Base progress report class.

    This base class implements the logic but no output.
    """
    def __init__(self, file=None):
        self._file = file or sys.stdout
        self.prev_progress = 0
        self.backwards_progress = False
        self.done = False
        self.last = 0

    def __call__(self, progress):
        assert progress >= 0.0

        # Cap progress at 1.0.
        if progress > 1.0:
            progress = 1.0
            LOG.debug("progress out of bounds: %.3f", progress)

        # Reset state on 0.0
        if progress == 0.0:
            self._start()

        # Check for backwards progress
        if progress < self.prev_progress:
            self.backwards_progress = True
        self.prev_progress = progress

        # print progress bar
        if not self.done:
            self._update(progress)

            # Finish on 1.0
            if progress >= 1.0:
                self.done = True
                self._finish()
                if self.backwards_progress:
                    LOG.debug("Progress went backwards!")

    def _start(self):
        self.prev_progress = 0
        self.backwards_progress = False
        self.done = False
        self.last = 0

    def _update(self, progress):
        raise NotImplementedError()

    def _finish(self):
        raise NotImplementedError()

class ProgressReportTTY(ProgressReport):
    """@brief Progress report subclass for TTYs.

    The progress bar is fully redrawn onscreen as progress is updated to give the
    impression of animation.
    """

    ## These width constants can't be changed yet without changing the code below to match.
    WIDTH = 50

    def _update(self, progress):
        self._file.write('\r')
        i = int(progress * self.WIDTH)
        self._file.write("[%-50s] %3d%%" % ('=' * i, round(progress * 100)))
        self._file.flush()

    def _finish(self):
        self._file.write("\n")

class ProgressReportNoTTY(ProgressReport):
    """@brief Progress report subclass for non-TTY output.

    A simpler progress bar is used than for the TTY version. Only the difference between
    the previous and current progress is drawn for each update, making the output suitable
    for piping to a file or similar output.
    """

    ## These width constants can't be changed yet without changing the code below to match.
    WIDTH = 40

    def _start(self):
        super(ProgressReportNoTTY, self)._start()

        self._file.write('[' + '---|' * 9 + '----]\n[')
        self._file.flush()

    def _update(self, progress):
        i = int(progress * self.WIDTH)
        delta = i - self.last
        self._file.write('=' * delta)
        self._file.flush()
        self.last = i

    def _finish(self):
        self._file.write("]\n")
        self._file.flush()

def print_progress(file=None):
    """@brief Progress printer factory.

    This factory function checks whether the output file is a TTY, and instantiates the
    appropriate subclass of ProgressReport.

    @param file The output file. Optional. If not provided, or if set to None, then sys.stdout
          will be used automatically.
    """

    if file is None:
        file = sys.stdout
    try:
        istty = os.isatty(file.fileno())
    except (OSError, AttributeError):
        # Either the file doesn't have a fileno method, or calling it returned an
        # error. In either case, just assume we're not connected to a TTY.
        istty = False

    klass = ProgressReportTTY if istty else ProgressReportNoTTY
    return klass(file)




# pyOCD debugger
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import (List, cast, TYPE_CHECKING)
import logging
import sys

from .base import SubcommandBase
from ..core.helpers import ConnectHelper
from ..core.target import Target
from ..utility.cmdline import (
    convert_session_options,
    convert_reset_type,
    int_base_0,
    )

if TYPE_CHECKING:
    from ..coresight.cortex_m import CortexM

LOG = logging.getLogger(__name__)

class ResetSubcommand(SubcommandBase):
    """@brief `pyocd reset` subcommand."""

    NAMES = ['reset']
    HELP = "Reset a target device."
    DEFAULT_LOG_LEVEL = logging.WARNING

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """@brief Add this subcommand to the subparsers object."""
        reset_parser = argparse.ArgumentParser(description='reset', add_help=False)

        reset_options = reset_parser.add_argument_group("reset options")
        reset_options.add_argument("-m", "--method", default=None, dest='reset_type', metavar="METHOD",
            help="Reset method to use (default, hw, sw, sysresetreq, vectreset, emulated). Takes precedence "
                 "over the 'reset_type' session option if set. If neither are set, the default reset type "
                 "from the core chosen with --core will be used (usually 'sw' but can be differ based "
                 "on the target type).")
        reset_options.add_argument("-c", "--core", default=0, type=int_base_0,
            help="Core number used to perform software reset. Only applies to software reset methods."
                 "Default is core 0.")
        reset_options.add_argument("-l", "--halt", action="store_true",
            help="Halt the core on the first instruction after reset. Defaults to disabled.")

        return [cls.CommonOptions.COMMON, cls.CommonOptions.CONNECT, reset_parser]

    def invoke(self) -> None:
        """@brief Handle 'reset' subcommand."""
        # Verify selected reset type, if set, before connecting.
        try:
            if self._args.reset_type is not None:
                _ = convert_reset_type(self._args.reset_type)
        except ValueError:
            LOG.error("Invalid reset method: %s", self._args.reset_type)
            return

        # Note that resume_on_disconnect is set to the inverse of the --halt argument. Obviously if
        # the target is resumed, halt won't stay in effect.
        session = ConnectHelper.session_with_chosen_probe(
                            project_dir=self._args.project_dir,
                            config_file=self._args.config,
                            user_script=self._args.script,
                            no_config=self._args.no_config,
                            pack=self._args.pack,
                            unique_id=self._args.unique_id,
                            target_override=self._args.target_override,
                            frequency=self._args.frequency,
                            blocking=(not self._args.no_wait),
                            connect_mode=self._args.connect_mode,
                            resume_on_disconnect=not self._args.halt,
                            reset_type=self._args.reset_type,
                            options=convert_session_options(self._args.options),
                            option_defaults=self._modified_option_defaults(),
                            )
        if session is None:
            LOG.error("No target device available to reset")
            sys.exit(1)
        try:
            # Get the reset type from the session option.
            try:
                the_reset_type = convert_reset_type(session.options.get('reset_type'))
            except ValueError:
                LOG.error("Invalid reset method: %s", session.options.get('reset_type'))
                return

            # Handle hw reset more efficiently using the probe directly, so we don't need can skip
            # discovery. However, if halting was requested we need full init even if performing a
            # hardware reset.
            is_hw_reset = (the_reset_type == Target.ResetType.HW) and not self._args.halt

            # Only init the board if performing a sw reset.
            session.open(init_board=(not is_hw_reset))
            assert session.probe
            assert session.target

            # If the reset type is default, get the concrete default type from the core so we can log it.
            if the_reset_type is None:
                session.target.selected_core = self._args.core

                # TODO This only works right now because all cores are CortexM. The default
                # reset type should really be moved to CoreTarget.
                the_reset_type = cast("CortexM", session.target.selected_core).default_reset_type

            LOG.info("Performing %s reset...", the_reset_type.name)
            if is_hw_reset:
                # For some probe types the probe still has to be connected to drive reset.
                session.probe.connect()
                session.probe.reset()
                session.probe.disconnect()
            else:
                if self._args.halt:
                    session.target.reset_and_halt(reset_type=the_reset_type)
                else:
                    session.target.reset(reset_type=the_reset_type)
            LOG.info("Done.")
        finally:
            session.close()




# pyOCD debugger
# Copyright (c) 2017-2020 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import os
from typing import (cast, Callable, Dict, IO, Iterable, List, Optional, Set, Tuple, Type, Union, TYPE_CHECKING)


from .cmsis_pack import (CmsisPack, CmsisPackDevice, MalformedCmsisPackError)
from .reset_sequence_maps import (RESET_SEQUENCE_TO_TYPE_MAP, RESET_TYPE_TO_SEQUENCE_MAP)
from ..family import FAMILIES
from .. import (normalise_target_type_name, TARGET)
from ...core import exceptions
from ...core.target import Target
from ...coresight.ap import APv1Address
from ...coresight.coresight_target import CoreSightTarget
from ...coresight.cortex_m import CortexM
from ...debug.sequences.delegates import DebugSequenceDelegate
from ...debug.sequences.functions import DebugSequenceCommonFunctions
from ...debug.sequences.sequences import (Block, DebugSequence, DebugSequenceExecutionContext)
from ...debug.sequences.scope import Scope
from ...debug.svd.loader import SVDFile
from ...core.session import Session
from ...probe.debug_probe import DebugProbe

if TYPE_CHECKING:
    from zipfile import ZipFile
    from cmsis_pack_manager import CmsisPackRef
    from ...utility.sequencer import CallSequence
    from ...core.core_target import CoreTarget
    from ...commands.execution_context import CommandSet
    from ...utility.notification import Notification

try:
    import cmsis_pack_manager
    CPM_AVAILABLE = True
except ImportError:
    CPM_AVAILABLE = False

LOG = logging.getLogger(__name__)

class ManagedPacksStub:
    @staticmethod
    def get_installed_packs(cache: Optional[object] = None) -> List:
        return []

    @staticmethod
    def get_installed_targets(cache: Optional[object] = None) -> List:
        return []

    @staticmethod
    def populate_target(device_name: str) -> None:
        pass

class ManagedPacksImpl:
    """@brief Namespace for managed CMSIS-Pack utilities.

    By managed, we mean managed by the cmsis-pack-manager package. All the methods on this class
    apply only to those packs managed by cmsis-pack-manager, not any targets from packs specified
    by the user.
    """

    @staticmethod
    def get_installed_packs(cache: Optional[cmsis_pack_manager.Cache] = None) -> List[CmsisPackRef]: # type:ignore
        """@brief Return a list containing CmsisPackRef objects for all installed packs."""
        if cache is None:
            cache = cmsis_pack_manager.Cache(True, True)
        results = []
        # packs_for_devices() returns only unique packs.
        for pack in cache.packs_for_devices(cache.index.values()):
            # Generate full path to the .pack file.
            pack_path = os.path.join(cache.data_path, pack.get_pack_name())

            # If the .pack file exists, the pack is installed.
            if os.path.isfile(pack_path):
                results.append(pack)
        return results

    @staticmethod
    def get_installed_targets(cache: Optional[cmsis_pack_manager.Cache] = None) -> List[CmsisPackDevice]: # type:ignore
        """@brief Return a list of CmsisPackDevice objects for installed pack targets."""
        if cache is None:
            cache = cmsis_pack_manager.Cache(True, True)
        results = []
        for pack in ManagedPacks.get_installed_packs(cache=cache):
            try:
                pack_path = os.path.join(cache.data_path, pack.get_pack_name())
                pack = CmsisPack(pack_path)
                results += list(pack.devices)
            except Exception as err:
                LOG.error("failure to access managed CMSIS-Pack: %s",
                        err, exc_info=Session.get_current().log_tracebacks)
        return sorted(results, key=lambda dev:dev.part_number)

    @staticmethod
    def populate_target(device_name: str) -> None:
        """@brief Add targets from cmsis-pack-manager matching the given name.

        Targets are added to the `#TARGET` list. A case-insensitive comparison against the
        device part number is used to find the target to populate. If multiple packs are installed
        that provide the same part numbers, all matching targets will be populated.
        """
        device_name = normalise_target_type_name(device_name)
        targets = ManagedPacks.get_installed_targets()
        for dev in targets:
            if device_name == normalise_target_type_name(dev.part_number):
                PackTargets.populate_device(dev)

if CPM_AVAILABLE:
    ManagedPacks = ManagedPacksImpl
else:
    ManagedPacks = ManagedPacksStub

class PackDebugSequenceDelegate(DebugSequenceDelegate):
    """! @brief Main delegate for debug sequences."""

    ## Map from pyocd reset types to the __connection variable reset type field.
    #
    # 0=error, 1=hw, 2=SYSRESETREQ, 3=VECTRESET
    RESET_TYPE_MAP = {
        Target.ResetType.HW: 1,
        Target.ResetType.SW: 2, # TODO pick default sw reset type
        Target.ResetType.SW_SYSRESETREQ: 2,
        Target.ResetType.SW_VECTRESET: 3,
        Target.ResetType.SW_EMULATED: 2, # no direct match
    }

    def __init__(self, target: CoreSightTarget, device: CmsisPackDevice) -> None:
        self._target = target
        self._session = target.session
        self._pack_device = device
        self._sequences: Set[DebugSequence] = device.sequences
        self._debugvars: Optional[Scope] = None
        self._functions = DebugSequenceCommonFunctions()

        self._session.options.subscribe(self._debugvars_did_change, 'pack.debug_sequences.debugvars')

    @property
    def all_sequences(self) -> Set[DebugSequence]:
        return self._sequences

    @property
    def cmsis_pack_device(self) -> CmsisPackDevice:
        """@brief Accessor for the pack device that contains the sequences."""
        return self._pack_device

    def get_root_scope(self, context: DebugSequenceExecutionContext) -> Scope:
        """@brief Return a scope that will be used as the parent of sequences."""
        # TODO should a fresh exec context be used? debugvars aren't supposed to depend on any
        # runtime debug settings nor do they even have access to those variables.
        if self._debugvars is not None:
            return self._debugvars

        # Populate a scope with definitions from the <debugvars> element, if defined. If not defined,
        # the the scope will be empty.
        self._debugvars = Scope(name="debugvars")
        debugvars_block = self._pack_device.debug_vars_sequence
        if debugvars_block is not None:
            # This is the only case where a Block will be pushed to the context stack.
            with context.push(debugvars_block, self._debugvars):
                debugvars_block.execute(context)

        # Now run the debugvars session option, if defined, as a block to override default
        # debugvars values from the <debugvars> element.
        debugvars_option = self._session.options.get('pack.debug_sequences.debugvars')
        if (debugvars_option is not None) and (debugvars_option.strip() != ""):
            debugvars_option_block = Block(debugvars_option)
            # This is the only case where a Block will be pushed to the context stack.
            with context.push(debugvars_option_block, self._debugvars):
                debugvars_option_block.execute(context)

        # Make all vars read-only.
        self._debugvars.freeze()

        # Debug log all debugvar values.
        if LOG.isEnabledFor(logging.INFO):
            for name in sorted(self._debugvars.variables):
                value = self._debugvars.get(name)
                LOG.info(f"debugvar '{name}' = {value:#x} ({value:d})")

        return self._debugvars

    def _debugvars_did_change(self, notification: Notification) -> None:
        """@brief Notification handler for change to pack.debug_sequences.debugvars option."""
        # Clear the cached debugvars scope to force it to be rebuilt.
        self._debugvars = None

    def _is_sequence_manually_disabled(self, name: str, pname: Optional[str] = None) -> bool:
        """@brief Check session options to see if the sequence has been disabled by the user."""
        disabled_seqs = self._session.options.get('pack.debug_sequences.disabled_sequences')
        if not disabled_seqs:
            return False

        name = name.casefold()
        if pname is not None:
            pname = pname.casefold()

        for dseq in disabled_seqs:
            if ':' in dseq:
                dseq, core_name = dseq.split(':')
                core_name = core_name.casefold()
            else:
                core_name = None
            dseq = dseq.casefold()

            if (name == dseq) and ((core_name is None) or (pname == core_name)):
                return True

        return False

    def run_sequence(self, name: str, pname: Optional[str] = None) -> Optional[Scope]:
        """@brief Run a top level debug sequence.

        @return The scope created while running the sequence is returned. If the sequence wasn't executed
            for some reason, e.g. it was disabled, then None is returned instead.
        """
        pname_desc = f" ({pname})" if (pname and LOG.isEnabledFor(logging.DEBUG)) else ""

        # Handle global debug sequence enable.
        if not self._session.options.get('pack.debug_sequences.enable'):
            LOG.debug("Not running debug sequence '%s'%s because all sequences are disabled",
                    name, pname_desc)
            return None

        # Error out for invalid sequence.
        if not self.has_sequence_with_name(name, pname):
            raise NameError(name)

        # Get the sequence object.
        seq = self.get_sequence_with_name(name, pname)

        # If the sequence is disabled, we won't run it.
        if not seq.is_enabled:
            LOG.debug("Not running disabled debug sequence '%s'%s", name, pname_desc)
            return None
        # Check for manual disabling of this sequence.
        if self._is_sequence_manually_disabled(name, pname):
            LOG.debug("Not running debug sequence '%s'%s because it was manually disabled", name, pname_desc)
            return None

        LOG.debug("Running debug sequence '%s'%s", name, pname_desc)

        # Create runtime context and contextified functions instance.
        context = DebugSequenceExecutionContext(self._session, self, pname)

        # Map optional pname to AP address. If the pname is not specified, then use the device's
        # first available AP. If no APs are known (eg haven't been discovered yet) then use 0.
        if pname:
            proc_map = self._pack_device.processors_map
            ap_address = proc_map[pname].ap_address
        else:
            ap = self._target.first_ap
            if ap is not None:
                ap_address = ap.address
            else:
                ap_address = APv1Address(0)

        # Set the default AP in the exec context.
        context.default_ap = ap_address

        # Activate the context while running this sequence, making the context available
        # to sequence functions.
        with context:
            try:
                executed_scope = seq.execute(context)
            except exceptions.Error as err:
                if pname:
                    LOG.error("Error while running debug sequence '%s' (core %s): %s", name, pname, err)
                else:
                    LOG.error("Error while running debug sequence '%s': %s", name, err)
                raise

        return executed_scope

    def sequences_for_pname(self, pname: Optional[str]) -> Dict[str, DebugSequence]:
        # Return *only* sequences with no Pname when passed pname=None. Otherwise we'd have
        # to mangle the dict keys to include pname since there can be multiple sequences with
        # the same name but different
        return {
            seq.name: seq
            for seq in self._sequences
            if (seq.pname is None) or (seq.pname == pname)
        }

    def has_sequence_with_name(self, name: str, pname: Optional[str] = None) -> bool:
        return name in self.sequences_for_pname(pname)

    def get_sequence_with_name(self, name: str, pname: Optional[str] = None) -> DebugSequence:
        return self.sequences_for_pname(pname)[name]

    def get_protocol(self) -> int:
        """@brief Return the value for the __protocol variable.
        __protocol fields:
        - [15:0] 0=error, 1=JTAG, 2=SWD, 3=cJTAG
        - [16] SWJ-DP present?
        - [17] switch through dormant state?
        """
        session = self._target.session
        assert session.probe, "must have a valid probe"
        # Not having a wire protocol set is allowed if performing pre-reset since it will only
        # execute ResetHardware (or equivalent), which can only access pins and such (theoretically).
        assert self._session.context_state.is_performing_pre_reset or session.probe.wire_protocol, \
            "must have valid, connected probe"
        if session.probe.wire_protocol == DebugProbe.Protocol.JTAG:
            protocol = 1
        elif session.probe.wire_protocol == DebugProbe.Protocol.SWD:
            protocol = 2
        else:
            protocol = 0 # Error
        if session.options.get('dap_swj_enable'):
            protocol |= 1 << 16
        if session.options.get('dap_swj_use_dormant'):
            protocol |= 1 << 17
        return protocol

    def get_connection_type(self) -> int:
        """@brief Return the value for the __connection variable.
        __connection fields:
        - [7:0] connection type: 0=error/disconnected, 1=for debug, 2=for flashing
        - [15:8] reset type: 0=error, 1=hw, 2=SYSRESETREQ, 3=VECTRESET
        - [16] connect under reset?
        - [17] pre-connect reset?
        """
        ctype = 1
        ctype |= self.RESET_TYPE_MAP.get(self._session.options.get('reset_type'), 0) << 8

        connect_mode = self._target.session.options.get('connect_mode')
        if connect_mode == 'under-reset':
            ctype |= 1 << 16

        # The pre-reset bit should only be set when running ResetHardware for a connect pre-reset.
        # This is stored in the is_performing_pre_reset session state variable, set by CoreSightTarget's
        # pre_connect() method.
        if self._session.context_state.is_performing_pre_reset:
            ctype |= 1 << 17
        return ctype

    def get_traceout(self) -> int:
        """@brief Return the value for the __traceout variable.
        __traceout fields:
        - [0] SWO enabled?
        - [1] parallel trace enabled?
        - [2] trace buffer enabled?
        - [21:16] selected parallel trace port size
        """
        # Set SWO bit depending on the option value.
        return 1 if self._target.session.options.get('enable_swv') else 0

    def get_sequence_functions(self) -> DebugSequenceCommonFunctions:
        return self._functions

class _PackTargetMethods:
    """@brief Container for methods added to the dynamically generated pack target subclass.

    We can't just make a subclass of CoreSightTarget out of these methods, because the superclass
    is variable based on the possible family class.
    """

    @staticmethod
    def _pack_target__init__(self, session: Session) -> None: # type:ignore
        """@brief Constructor for dynamically created target class."""
        super(self.__class__, self).__init__(session, self._pack_device.memory_map)

        self.vendor = self._pack_device.vendor
        self.part_families = self._pack_device.families
        self.part_number = self._pack_device.part_number

        self._svd_location = SVDFile(filename=self._pack_device.svd)

        self.debug_sequence_delegate = PackDebugSequenceDelegate(self, self._pack_device)

    @staticmethod
    def _pack_target_create_init_sequence(self) -> CallSequence: # type:ignore
        """@brief Creates an init task to set the default reset type."""
        seq = super(self.__class__, self).create_init_sequence()

        seq.wrap_task('discovery',
            lambda seq: seq.insert_after('create_cores',
                            ('configure_core_reset', self.configure_core_reset)
                            )
            )
        return seq

    @staticmethod
    def _pack_target_configure_core_reset(self) -> None: # type:ignore
        """@brief Init sequence method to configure resets for all cores.

        This init sequence method is designed to run after the cores have been created by standard
        discovery.

        Sets each core's default reset type to the one specified in the pack, and configures the
        list of enabled reset types.
        """
        for core_num, core in self.cores.items():
            # Look up the processor info for this core.
            core_ap_addr = core.ap.address
            try:
                proc_info = self._pack_device.processors_ap_map[core_ap_addr]
            except KeyError:
                LOG.debug("core #%d not specified in DFP", core_num)
                continue

            # Get this processor's list of sequences.
            sequences = self.debug_sequence_delegate.sequences_for_pname(proc_info.name)

            def is_reset_sequence_enabled(name: str) -> bool:
                return (name not in sequences) or sequences[name].is_enabled

            # Set the supported reset types by filtering existing supported reset types.
            updated_reset_types: Set[Target.ResetType] = set()
            for resettype in core._supported_reset_types:
                # These two types are not in the map, and should always be present.
                if resettype in (Target.ResetType.SW, Target.ResetType.SW_EMULATED):
                    updated_reset_types.add(resettype)
                    continue

                resettype_sequence_name = RESET_TYPE_TO_SEQUENCE_MAP[resettype]
                if is_reset_sequence_enabled(resettype_sequence_name):
                    updated_reset_types.add(resettype)

            # Special case to enable processor reset even when the core doesn't support VECTRESET, if
            # there is a non-default ResetProcessor sequence definition.
            if ((Target.ResetType.SW_CORE not in updated_reset_types) # type:ignore
                    and ('ResetProcessor' in sequences)
                    and sequences['ResetProcessor'].is_enabled):
                updated_reset_types.add(Target.ResetType.SW_CORE) # type:ignore

            core._supported_reset_types = updated_reset_types
            LOG.debug(f"updated DFP core #{core_num} reset types: {core._supported_reset_types}")

            default_reset_seq = proc_info.default_reset_sequence

            # Check that the default reset sequence is a standard sequence. The specification allows for
            # custom reset sequences to be used, but that is not supported by pyocd yet.
            # TODO support custom default reset sequences (requires a new reset type)
            if default_reset_seq not in RESET_SEQUENCE_TO_TYPE_MAP:
                if default_reset_seq in sequences:
                    # Custom reset sequence, not yet supported by pyocd.
                    LOG.warning("DFP device definition error: custom reset sequences are not yet supported "
                                "by pyocd; core #%d (%s) requested default reset sequence %s",
                                core_num, proc_info.name, default_reset_seq)
                else:
                    # Invalid/unknown default reset sequence.
                    LOG.warning("DFP device definition error: specified default reset sequence %s "
                                "for core #%d (%s) does not exist",
                                default_reset_seq, core_num, proc_info.name)

            # Handle multicore debug mode causing secondary cores to default to processor reset.
            did_force_core_reset = False
            if (self.session.options.get('enable_multicore_debug')
                    and (core_num != self.session.options.get('primary_core'))):
                if not is_reset_sequence_enabled('ResetProcessor'):
                    LOG.warning("Multicore debug mode cannot select processor reset for secondary core "
                                "#%d (%s) because it is disabled by the DFP; using emulated processor "
                                "reset instead", core_num, proc_info.name)
                    core.default_reset_type = Target.ResetType.SW_EMULATED
                    continue
                else:
                    default_reset_seq = 'ResetProcessor'
                    did_force_core_reset = True

            # Verify that the specified default reset sequence hasn't been disabled.
            if not is_reset_sequence_enabled(default_reset_seq):
                # Only log a warning if we didn't decide to use core reset due to multicore mode.
                if not did_force_core_reset:
                    LOG.warning("DFP device definition conflict: specified default reset sequence %s "
                            "for core #%d (%s) is disabled by the DFP",
                            default_reset_seq, core_num, proc_info.name)

                # Map from disabled default to primary and secondary fallbacks.
                RESET_FALLBACKS: Dict[str, Tuple[str, str]] = {
                    'ResetSystem':      ('ResetProcessor', 'ResetHardware'),
                    'ResetHardware':    ('ResetSystem', 'ResetProcessor'),
                    'ResetProcessor':   ('ResetSystem', 'ResetHardware'),
                }

                # Select another default.
                fallbacks = RESET_FALLBACKS[default_reset_seq]
                if is_reset_sequence_enabled(fallbacks[0]):
                    default_reset_seq = fallbacks[0]
                elif is_reset_sequence_enabled(fallbacks[1]):
                    default_reset_seq = fallbacks[1]
                else:
                    LOG.warning("DFP device definition conflict: all reset types are disabled for "
                            "core #%d (%s) by the DFP; using emulated core reset",
                            default_reset_seq, core_num)
                    core.default_reset_type = Target.ResetType.SW_EMULATED
                    continue

            LOG.info("Setting core #%d (%s) default reset sequence to %s",
                    core_num, proc_info.name, default_reset_seq)
            core.default_reset_type = RESET_SEQUENCE_TO_TYPE_MAP[default_reset_seq]

    @staticmethod
    def _pack_target_add_core(_self, core: CoreTarget) -> None:
        """@brief Override to set node name of added core to its pname."""
        pname = _self._pack_device.processors_ap_map[cast(CortexM, core).ap.address].name
        core.node_name = pname
        CoreSightTarget.add_core(_self, core)

    @staticmethod
    def _pack_target_add_target_command_groups(_self, command_set: CommandSet):
        """@brief Add pack related commands to the command set."""
        command_set.add_command_group('pack-target')


class PackTargets:
    """@brief Namespace for CMSIS-Pack target generation utilities. """

    @staticmethod
    def _find_family_class(dev: CmsisPackDevice) -> Type[CoreSightTarget]:
        """@brief Search the families list for matching entry."""
        for familyInfo in FAMILIES:
            # Skip if wrong vendor.
            if dev.vendor != familyInfo.vendor:
                continue

            # Scan each level of family plus part number, from specific to generic.
            for compare_name in reversed(dev.families + [dev.part_number]):
                # Require the regex to match the entire family name.
                match = familyInfo.matches.match(compare_name)
                if match and match.span() == (0, len(compare_name)):
                    LOG.debug("using family class %s for %s (matched against %s)",
                            familyInfo.klass.__name__, dev.part_number, compare_name)
                    return familyInfo.klass

        # Didn't match, so return default target superclass.
        return CoreSightTarget

    @staticmethod
    def _generate_pack_target_class(dev: CmsisPackDevice) -> Optional[type]:
        """@brief Generates a new target class from a CmsisPackDevice.

        @param dev A CmsisPackDevice object.
        @return A new subclass of either CoreSightTarget or one of the family classes.
        """
        try:
            # Look up the target family superclass.
            superklass = PackTargets._find_family_class(dev)

            # Replace spaces and dashes with underscores on the new target subclass name.
            subclassName = normalise_target_type_name(dev.part_number).capitalize()

            # Create a new subclass for this target.
            targetClass = type(subclassName, (superklass,), {
                        "_pack_device": dev,
                        "__init__": _PackTargetMethods._pack_target__init__,
                        "create_init_sequence": _PackTargetMethods._pack_target_create_init_sequence,
                        "configure_core_reset": _PackTargetMethods._pack_target_configure_core_reset,
                        "add_core": _PackTargetMethods._pack_target_add_core,
                        "add_target_command_groups": _PackTargetMethods._pack_target_add_target_command_groups,
                    })
            return targetClass
        except (MalformedCmsisPackError, FileNotFoundError) as err:
            LOG.warning(err)
            return None

    @staticmethod
    def populate_device(dev: CmsisPackDevice) -> None:
        """@brief Generates and populates the target defined by a CmsisPackDevice.

        The new target class is added to the `#TARGET` list.

        @param dev A CmsisPackDevice object.
        """
        try:
            # Check if we're even going to populate this target before bothing to build the class.
            part = normalise_target_type_name(dev.part_number)
            if part in TARGET:
                LOG.debug("did not populate target for DFP part number %s because there is already "
                        "a %s target installed", dev.part_number, part)
                return

            # Generate target subclass and install it.
            tgt = PackTargets._generate_pack_target_class(dev)
            if tgt:
                TARGET[part] = tgt
        except (MalformedCmsisPackError, FileNotFoundError) as err:
            LOG.warning(err)

    PackReferenceType = Union[CmsisPack, str, "ZipFile", IO[bytes]]

    @staticmethod
    def process_targets_from_pack(
            pack_list: Union[PackReferenceType, Iterable[PackReferenceType]],
            cb: Callable[[CmsisPackDevice], None]
        ) -> None:
        """@brief Invoke a callable on devices defined in the provided CMSIS-Pack(s).

        @param pack_list Sequence of strings that are paths to .pack files, file objects,
            ZipFile instances, or CmsisPack instance. May also be a single object of one of
            the accepted types.
        @param cb Callable to run. Must take a CmsisPackDevice object as the sole parameter and return None.
        """
        if not isinstance(pack_list, (list, tuple)):
            pack_list = [pack_list] # type:ignore
        for pack_or_path in pack_list: # type:ignore
            if isinstance(pack_or_path, CmsisPack):
                pack = pack_or_path
            else:
                pack = CmsisPack(pack_or_path)
            for dev in pack.devices:
                cb(dev)

    @staticmethod
    def populate_targets_from_pack(pack_list: Union[PackReferenceType, Iterable[PackReferenceType]]) -> None:
        """@brief Adds targets defined in the provided CMSIS-Pack.

        Targets are added to the `#TARGET` list.

        @param pack_list Sequence of strings that are paths to .pack files, file objects,
            ZipFile instances, or CmsisPack instance. May also be a single object of one of
            the accepted types.
        """
        PackTargets.process_targets_from_pack(pack_list, PackTargets.populate_device)

def is_pack_target_available(target_name: str, session: Session) -> bool:
    """@brief Test whether a given target type is available."""
    # Create targets from provided CMSIS pack.
    if session.options['pack'] is not None:
        target_types = []
        def collect_target_type(dev: CmsisPackDevice) -> None:
            part = normalise_target_type_name(dev.part_number)
            target_types.append(part)
        PackTargets.process_targets_from_pack(session.options['pack'], collect_target_type)
        return target_name.lower() in target_types

    # Check whether a managed pack contains the target.
    return any(
                (target_name.lower() == dev.part_number.lower())
                for dev in ManagedPacks.get_installed_targets()
                )





# pyOCD debugger
# Copyright (c) 2006-2021 Arm Limited
# Copyright (c) 2020 Patrick Huesmann
# Copyright (c) 2021 mentha
# Copyright (c) 2021-2023 Chris Reed
# Copyright (c) 2022 Harper Weigle
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading
import platform
import errno
import queue
from typing import Optional

from .interface import Interface
from .common import (
    USB_CLASS_HID,
    filter_device_by_class,
    is_known_device_string,
    is_known_cmsis_dap_vid_pid,
    generate_device_unique_id,
    )
from ..dap_access_api import DAPAccessIntf
from ....utility.timeout import Timeout

LOG = logging.getLogger(__name__)
TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

try:
    import usb.core
    import usb.util
    try:
        from libusb_package import find as usb_find
    except ImportError:
        from usb.core import find as usb_find
except ImportError:
    IS_AVAILABLE = False
else:
    IS_AVAILABLE = True


class PyUSB(Interface):
    """@brief CMSIS-DAP USB interface class using pyusb for the backend."""

    isAvailable = IS_AVAILABLE

    did_show_no_libusb_warning = False

    def __init__(self, dev):
        super().__init__()
        self.vid = dev.idVendor
        self.pid = dev.idProduct
        self.product_name = dev.product or f"{dev.idProduct:#06x}"
        self.vendor_name = dev.manufacturer or f"{dev.idVendor:#06x}"
        self.serial_number = dev.serial_number \
                or generate_device_unique_id(dev.idProduct, dev.idVendor, dev.bus, dev.address)
        self.ep_out = None
        self.ep_in = None
        self.dev = None
        self.intf_number = None
        self.kernel_driver_was_attached = False
        self.closed = True
        self.thread = None
        self.rx_stop_event = threading.Event()
        self.rcv_data: queue.SimpleQueue[bytes] = queue.SimpleQueue()
        self.read_sem = threading.Semaphore(0)
        self._read_thread_did_exit: bool = False
        self._read_thread_exception: Optional[Exception] = None

    def open(self):
        assert self.closed is True

        # Get device handle
        dev = usb_find(custom_match=FindDap(self.serial_number))
        if dev is None:
            raise DAPAccessIntf.DeviceError(f"Probe {self.serial_number} not found")

        # get active config
        config = dev.get_active_configuration()

        # Get count of HID interfaces and create the matcher object
        hid_interface_count = len(list(usb.util.find_descriptor(config, find_all=True, bInterfaceClass=USB_CLASS_HID)))
        matcher = MatchCmsisDapv1Interface(hid_interface_count)

        # Get CMSIS-DAPv1 interface
        interface = usb.util.find_descriptor(config, custom_match=matcher)
        if interface is None:
            raise DAPAccessIntf.DeviceError(f"Probe {self.serial_number} has no CMSIS-DAPv1 interface")
        interface_number = interface.bInterfaceNumber

        # Find endpoints
        ep_in, ep_out = None, None
        for endpoint in interface:
            if endpoint.bEndpointAddress & usb.util.ENDPOINT_IN:
                ep_in = endpoint
            else:
                ep_out = endpoint

        # Detach kernel driver
        self.kernel_driver_was_attached = False
        try:
            if dev.is_kernel_driver_active(interface_number):
                LOG.debug("Detaching Kernel Driver of Interface %d from USB device (VID=%04x PID=%04x).", interface_number, dev.idVendor, dev.idProduct)
                dev.detach_kernel_driver(interface_number)
                self.kernel_driver_was_attached = True
        except usb.core.USBError as e:
            LOG.warning("USB Kernel Driver Detach Failed ([%s] %s). Attached driver may interfere with pyOCD operations.", e.errno, e.strerror)
        except NotImplementedError as e:
            # Some implementations don't don't have kernel attach/detach
            LOG.debug("Probe %s: USB kernel driver detaching is not supported. Attached HID driver may interfere with pyOCD operations.", self.serial_number)

        # Explicitly claim the interface
        try:
            usb.util.claim_interface(dev, interface_number)
        except usb.core.USBError as exc:
            raise DAPAccessIntf.DeviceError(f"Unable to claim interface for probe {self.serial_number}") from exc

        # Update all class variables if we made it here
        self.ep_out = ep_out
        self.ep_in = ep_in
        self.dev = dev
        self.intf_number = interface_number

        # Start RX thread as the last step
        self.closed = False
        self.start_rx()

    def start_rx(self):
        # Flush the RX buffers by reading until timeout exception
        try:
            while True:
                self.ep_in.read(self.ep_in.wMaxPacketSize, 1)
        except usb.core.USBError:
            # USB timeout expected
            pass

        # Start RX thread
        thread_name = f"CMSIS-DAPv1 receive ({self.serial_number})"
        self.thread = threading.Thread(target=self.rx_task, name=thread_name)
        self.thread.daemon = True
        self.thread.start()

    def rx_task(self):
        try:
            while not self.rx_stop_event.is_set():
                self.read_sem.acquire()
                if not self.rx_stop_event.is_set():
                    read_data = self.ep_in.read(self.ep_in.wMaxPacketSize,
                            timeout=self.DEFAULT_USB_TIMEOUT_MS).tobytes()

                    # This trace log is commented out to reduce clutter, but left in to leave available
                    # when debugging rx_task issues.
                    # if TRACE.isEnabledFor(logging.DEBUG):
                    #     # Strip off trailing zero bytes to reduce clutter.
                    #     TRACE.debug("  USB IN < (%d) %s", len(read_data),
                    #                 ' '.join([f'{i:02x}' for i in read_data.rstrip(b'\x00')]))

                    self.rcv_data.put(read_data)
        except Exception as err:
            TRACE.debug("rx_task exception: %s", err)
            self._read_thread_exception = err
        finally:
            self._swo_thread_did_exit = True

    @staticmethod
    def get_all_connected_interfaces():
        """@brief Returns all the connected CMSIS-DAP devices.

        returns an array of PyUSB (Interface) objects
        """
        # find all cmsis-dap devices
        try:
            all_devices = usb_find(find_all=True, custom_match=FindDap())
        except usb.core.NoBackendError:
            if not PyUSB.did_show_no_libusb_warning:
                LOG.warning("CMSIS-DAPv1 probes may not be detected because no libusb library was found.")
                PyUSB.did_show_no_libusb_warning = True
            return []

        # iterate on all devices found
        boards = [PyUSB(board) for board in all_devices]

        return boards

    def write(self, data):
        """@brief Write data on the OUT endpoint associated to the HID interface"""

        report_size = self.packet_size
        if self.ep_out:
            report_size = self.ep_out.wMaxPacketSize

        # Trace output data before padding.
        if TRACE.isEnabledFor(logging.DEBUG):
            TRACE.debug("  USB OUT> (%d) %s", len(data), ' '.join([f'{i:02x}' for i in data]))

        data.extend([0] * (report_size - len(data)))

        self.read_sem.release()

        if not self.ep_out:
            bmRequestType = 0x21       # Host to device request of type Class of Recipient Interface
            bmRequest = 0x09           # Set_REPORT (HID class-specific request for transferring data over EP0)
            wValue = 0x200             # Issuing an OUT report
            wIndex = self.intf_number  # interface number for HID
            self.dev.ctrl_transfer(bmRequestType, bmRequest, wValue, wIndex, data,
                    timeout=self.DEFAULT_USB_TIMEOUT_MS)
        else:
            self.ep_out.write(data, timeout=self.DEFAULT_USB_TIMEOUT_MS)

    def read(self):
        """@brief Read data on the IN endpoint associated to the HID interface"""
        if self.closed:
            return b''
        elif self._read_thread_did_exit:
            raise DAPAccessIntf.DeviceError(f"Probe {self.serial_number} read thread exited unexpectedly") \
                from self._read_thread_exception

        try:
            data = self.rcv_data.get(True, self.DEFAULT_USB_TIMEOUT_S)
        except queue.Empty:
            raise DAPAccessIntf.DeviceError(f"Timeout reading from probe {self.serial_number}") from None

        # Trace when the higher layer actually gets a packet previously read.
        if TRACE.isEnabledFor(logging.DEBUG):
            # Strip off trailing zero bytes to reduce clutter.
            TRACE.debug("  USB RD < (%d) %s", len(data),
                    ' '.join([f'{i:02x}' for i in bytes(data).rstrip(b'\x00')]))

        return data

    def close(self):
        """@brief Close the interface"""
        assert self.closed is False

        LOG.debug("closing interface")
        self.closed = True
        self.rx_stop_event.set()
        self.read_sem.release()
        self.thread.join()
        self.rx_stop_event.clear() # Reset the stop event.
        self.rcv_data = queue.SimpleQueue() # Recreate queue to ensure it's empty.
        usb.util.release_interface(self.dev, self.intf_number)
        if self.kernel_driver_was_attached:
            try:
                self.dev.attach_kernel_driver(self.intf_number)
            except Exception as exception:
                LOG.warning('Exception attaching kernel driver: %s',
                                str(exception))
        usb.util.dispose_resources(self.dev)
        self.ep_out = None
        self.ep_in = None
        self.dev = None
        self.intf_number = None
        self.kernel_driver_was_attached = False
        self.thread = None
        self._read_thread_did_exit = False
        self._read_thread_exception = None

class MatchCmsisDapv1Interface:
    """@brief Match class for finding CMSIS-DAPv1 interface.

    This match class performs several tests on the provided USB interface descriptor, to
    determine whether it is a CMSIS-DAPv1 interface. These requirements must be met by the
    interface:

    1. If there is more than one HID interface on the device, the interface must have an interface
        name string containing "CMSIS-DAP".
    2. bInterfaceClass must be 0x03 (HID).
    3. bInterfaceSubClass must be 0.
    4. Must have interrupt in endpoint, with an optional interrupt out endpoint, in that order.
    """

    def __init__(self, hid_interface_count):
        """@brief Constructor."""
        self._hid_count = hid_interface_count

    def __call__(self, interface):
        """@brief Return True if this is a CMSIS-DAPv1 interface."""
        try:
            if self._hid_count > 1:
                interface_name = usb.util.get_string(interface.device, interface.iInterface)

                # This tells us whether the interface is CMSIS-DAP, but not whether it's v1 or v2.
                if (interface_name is None) or ("CMSIS-DAP" not in interface_name):
                    return False

            # Now check the interface class to distinguish v1 from v2.
            if (interface.bInterfaceClass != USB_CLASS_HID) \
                or (interface.bInterfaceSubClass != 0):
                return False

            # Must have either 1 or 2 endpoints.
            if interface.bNumEndpoints not in (1, 2):
                return False

            endpoint_attrs = [
                (usb.util.endpoint_direction(ep.bEndpointAddress),
                 usb.util.endpoint_type(ep.bmAttributes))
                 for ep in interface
            ]

            # Possible combinations of endpoints
            ENDPOINT_ATTRS_ALLOWED = [
                # One interrupt endpoint IN
                [(usb.util.ENDPOINT_IN, usb.util.ENDPOINT_TYPE_INTR)],
                # Two interrupt endpoints, first one IN, second one OUT
                [(usb.util.ENDPOINT_IN, usb.util.ENDPOINT_TYPE_INTR),
                 (usb.util.ENDPOINT_OUT, usb.util.ENDPOINT_TYPE_INTR)],
                # Two interrupt endpoints, first one OUT, second one IN
                [(usb.util.ENDPOINT_OUT, usb.util.ENDPOINT_TYPE_INTR),
                 (usb.util.ENDPOINT_IN, usb.util.ENDPOINT_TYPE_INTR)],
            ]
            if endpoint_attrs not in ENDPOINT_ATTRS_ALLOWED:
                return False

            # All checks passed, this is a CMSIS-DAPv2 interface!
            return True

        except (UnicodeDecodeError, IndexError):
            # UnicodeDecodeError exception can be raised if the device has a corrupted interface name.
            # Certain versions of STLinkV2 are known to have this problem. If we can't read the
            # interface name, there's no way to tell if it's a CMSIS-DAPv2 interface.
            #
            # IndexError can be raised if an endpoint is missing.
            return False

class FindDap:
    """@brief CMSIS-DAP match class to be used with usb.core.find"""

    def __init__(self, serial=None):
        """@brief Create a new FindDap object with an optional serial number"""
        self._serial = serial

    def __call__(self, dev):
        """@brief Return True if this is a DAP device, False otherwise"""
        # Check if the device class is a valid one for CMSIS-DAP.
        if filter_device_by_class(dev.idVendor, dev.idProduct, dev.bDeviceClass):
            return False

        known_cmsis_dap = is_known_cmsis_dap_vid_pid(dev.idVendor, dev.idProduct)
        try:
            # First attempt to get the active config. This produces a more direct error
            # when you don't have device permissions on Linux
            config = dev.get_active_configuration()

            # Now read the product name string.
            device_string = dev.product
            if ((device_string is None) or (not is_known_device_string(device_string))) and (not known_cmsis_dap):
                return False

            # Get count of HID interfaces.
            hid_interface_count = len(list(usb.util.find_descriptor(config, find_all=True, bInterfaceClass=USB_CLASS_HID)))

            # Find the CMSIS-DAPv1 interface.
            matcher = MatchCmsisDapv1Interface(hid_interface_count)
            cmsis_dap_interface = usb.util.find_descriptor(config, custom_match=matcher)
        except usb.core.USBError as error:
            if error.errno == errno.EACCES and platform.system() == "Linux":
                msg = ("%s while trying to interrogate a USB device "
                   "(VID=%04x PID=%04x). This can probably be remedied with a udev rule. "
                   "See <https://github.com/pyocd/pyOCD/tree/master/udev> for help." %
                   (error, dev.idVendor, dev.idProduct))
                # If we recognize this device as one that should be CMSIS-DAP, we can raise
                # the level of the log message since it's almost certainly a permissions issue.
                if known_cmsis_dap:
                    LOG.warning(msg)
                else:
                    LOG.debug(msg)
            else:
                LOG.debug("Error accessing USB device (VID=%04x PID=%04x): %s",
                    dev.idVendor, dev.idProduct, error)
            return False
        except (IndexError, NotImplementedError, ValueError, UnicodeDecodeError) as error:
            LOG.debug("Error accessing USB device (VID=%04x PID=%04x): %s", dev.idVendor, dev.idProduct, error)
            return False

        if cmsis_dap_interface is None:
            return False
        if self._serial is not None:
            if dev.serial_number is None:
                if self._serial == "":
                    return True
                if self._serial == generate_device_unique_id(dev.idProduct, dev.idVendor, dev.bus, dev.address):
                    return True
            if self._serial != dev.serial_number:
                return False
        return True



# pyOCD debugger
# Copyright (c) 2015-2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .fpb import HardwareBreakpoint
from ..core.target import Target
from .component import CoreSightComponent
import logging

LOG = logging.getLogger(__name__)

# Need a local copy to prevent circular import.
# Debug Exception and Monitor Control Register
DEMCR = 0xE000EDFC
# DWTENA in armv6 architecture reference manual
DEMCR_TRCENA = (1 << 24)

class Watchpoint(HardwareBreakpoint):
    def __init__(self, comp_register_addr, provider):
        super(Watchpoint, self).__init__(comp_register_addr, provider)
        self.addr = 0
        self.size = 0
        self.func = 0

class DWT(CoreSightComponent):
    """@brief Data Watchpoint and Trace version 1.0"""

    # DWT registers
    #
    # The addresses are offsets from the base address.
    DWT_CTRL = 0x00000000
    DWT_CYCCNT = 0x00000004
    DWT_CPICNT = 0x00000008
    DWT_EXCCNT = 0x0000000C
    DWT_SLEEPCNT = 0x00000010
    DWT_LSUCNT = 0x00000014
    DWT_FOLDCNT = 0x00000018
    DWT_PCSR = 0x0000001C
    DWT_COMP_BASE = 0x00000020
    DWT_MASK_OFFSET = 4
    DWT_FUNCTION_OFFSET = 8
    DWT_COMP_BLOCK_SIZE = 0x10

    DWT_CTRL_NUM_COMP_MASK = (0xF << 28)
    DWT_CTRL_NUM_COMP_SHIFT = 28
    DWT_CTRL_CYCEVTENA_MASK = (1 << 22)
    DWT_CTRL_FOLDEVTENA_MASK = (1 << 21)
    DWT_CTRL_LSUEVTENA_MASK = (1 << 20)
    DWT_CTRL_SLEEPEVTENA_MASK = (1 << 19)
    DWT_CTRL_EXCEVTENA_MASK = (1 << 18)
    DWT_CTRL_CPIEVTENA_MASK = (1 << 17)
    DWT_CTRL_EXCTRCENA_MASK = (1 << 16)
    DWT_CTRL_PCSAMPLENA_MASK = (1 << 12)
    DWT_CTRL_SYNCTAP_MASK = (0x3 << 10)
    DWT_CTRL_SYNCTAP_SHIFT = 10
    DWT_CTRL_CYCTAP_MASK = (1 << 9)
    DWT_CTRL_POSTINIT_MASK = (0xF << 5)
    DWT_CTRL_POSTINIT_SHIFT = 5
    DWT_CTRL_POSTRESET_MASK = (0xF << 1)
    DWT_CTRL_POSTRESET_SHIFT = 1
    DWT_CTRL_CYCCNTENA_MASK = (1 << 0)

    WATCH_TYPE_TO_FUNCT = {
                            Target.WatchpointType.READ: 5,
                            Target.WatchpointType.WRITE: 6,
                            Target.WatchpointType.READ_WRITE: 7,
                            5: Target.WatchpointType.READ,
                            6: Target.WatchpointType.WRITE,
                            7: Target.WatchpointType.READ_WRITE,
                            }

    # Only sizes that are powers of 2 are supported
    # Breakpoint size = MASK**2
    WATCH_SIZE_TO_MASK = dict((2**i, i) for i in range(0,32))

    def __init__(self, ap, cmpid=None, addr=None):
        super(DWT, self).__init__(ap, cmpid, addr)
        self.watchpoints = []
        self.watchpoint_used = 0
        self.dwt_configured = False

    @property
    def watchpoint_count(self):
        return len(self.watchpoints)

    def init(self):
        """@brief Inits the DWT.

        Reads the number of hardware watchpoints available on the core and makes sure that they
        are all disabled and ready for future use.
        """
        # Make sure trace is enabled.
        demcr = self.ap.read_memory(DEMCR)
        if (demcr & DEMCR_TRCENA) == 0:
            demcr |= DEMCR_TRCENA
            self.ap.write_memory(DEMCR, demcr)

        dwt_ctrl = self.ap.read_memory(self.address + self.DWT_CTRL)
        watchpoint_count = (dwt_ctrl & self.DWT_CTRL_NUM_COMP_MASK) >> self.DWT_CTRL_NUM_COMP_SHIFT
        LOG.info("%d hardware watchpoints", watchpoint_count)
        for i in range(watchpoint_count):
            comparatorAddress = self.address + self.DWT_COMP_BASE + self.DWT_COMP_BLOCK_SIZE * i
            self.watchpoints.append(Watchpoint(comparatorAddress, self))
            self.ap.write_memory(comparatorAddress + self.DWT_FUNCTION_OFFSET, 0)

        # Enable cycle counter.
        self.ap.write32(self.address + self.DWT_CTRL, self.DWT_CTRL_CYCCNTENA_MASK)
        self.dwt_configured = True

    def find_watchpoint(self, addr, size, type):
        for watch in self.watchpoints:
            if watch.addr == addr and watch.size == size and watch.func == self.WATCH_TYPE_TO_FUNCT[type]:
                return watch
        return None

    def set_watchpoint(self, addr, size, type):
        """@brief Set a hardware watchpoint."""
        if self.dwt_configured is False:
            self.init()

        watch = self.find_watchpoint(addr, size, type)
        if watch is not None:
            return True

        if type not in self.WATCH_TYPE_TO_FUNCT:
            LOG.error("Invalid watchpoint type %i", type)
            return False

        for watch in self.watchpoints:
            if watch.func == 0:
                watch.addr = addr
                watch.func = self.WATCH_TYPE_TO_FUNCT[type]
                watch.size = size

                if size not in self.WATCH_SIZE_TO_MASK:
                    LOG.error('Watchpoint of size %d not supported by device', size)
                    return False

                mask = self.WATCH_SIZE_TO_MASK[size]
                self.ap.write_memory(watch.comp_register_addr + self.DWT_MASK_OFFSET, mask)
                if self.ap.read_memory(watch.comp_register_addr + self.DWT_MASK_OFFSET) != mask:
                    LOG.error('Watchpoint of size %d not supported by device', size)
                    return False

                self.ap.write_memory(watch.comp_register_addr, addr)
                self.ap.write_memory(watch.comp_register_addr + self.DWT_FUNCTION_OFFSET, watch.func)
                self.watchpoint_used += 1
                return True

        LOG.error('No more watchpoints are available, dropped watchpoint at 0x%08x', addr)
        return False

    def remove_watchpoint(self, addr, size=None, type=None):
        """@brief Remove a hardware watchpoint.

        size and type are optional. If not specified, all watchpoints matching other parameters, or only
        the address if neither size nor type are specified, will be removed.
        """
        for watch in self.watchpoints:
            if (watch.addr == addr
                    and (size is None or watch.size == size)
                    and (type is None or watch.func == self.WATCH_TYPE_TO_FUNCT[type])):
                watch.func = 0
                self.ap.write_memory(watch.comp_register_addr + self.DWT_FUNCTION_OFFSET, 0)
                self.watchpoint_used -= 1

    def remove_all_watchpoints(self):
        for watch in self.watchpoints:
            if watch.func != 0:
                self.remove_watchpoint(watch.addr, watch.size, self.WATCH_TYPE_TO_FUNCT[watch.func])

    def get_watchpoints(self):
        return [watch for watch in self.watchpoints if watch.func != 0]

    @property
    def cycle_count(self):
        return self.ap.read32(self.address + self.DWT_CYCCNT)

    @cycle_count.setter
    def cycle_count(self, value):
        self.ap.write32(self.address + self.DWT_CYCCNT, value)

class DWTv2(DWT):
    """@brief Data Watchpoint and Trace version 2.x

    This version is present in v8-M platforms.

    - DWT 2.0 appears in v8.0-M
    - DWT 2.1 appears in v8.1-M and adds the VMASKn registers.
    """

    DWT_ACTION_DEBUG_EVENT = 0x00000010

    ## Map from watchpoint type to FUNCTIONn.MATCH field value.
    WATCH_TYPE_TO_FUNCT = {
                            Target.WatchpointType.READ: 0b0110,
                            Target.WatchpointType.WRITE: 0b0101,
                            Target.WatchpointType.READ_WRITE: 0b0100,
                            0b0110: Target.WatchpointType.READ,
                            0b0101: Target.WatchpointType.WRITE,
                            0b0100: Target.WatchpointType.READ_WRITE,
                            }

    ## Map from data access size to pre-shifted DATAVSIZE field value.
    DATAVSIZE_MAP = {
                        1: (0 << 10),
                        2: (1 << 10),
                        4: (2 << 10),
                    }

    def set_watchpoint(self, addr, size, type):
        """@brief Set a hardware watchpoint."""
        if self.dwt_configured is False:
            self.init()

        watch = self.find_watchpoint(addr, size, type)
        if watch is not None:
            return True

        if type not in self.WATCH_TYPE_TO_FUNCT:
            LOG.error("Invalid watchpoint type %i", type)
            return False

        # Only support sizes that can be handled with a single comparator.
        if size not in (1, 2, 4):
            LOG.error("Invalid watchpoint size %d", size)
            return False

        for watch in self.watchpoints:
            if watch.func == 0:
                watch.addr = addr
                watch.func = self.WATCH_TYPE_TO_FUNCT[type]
                watch.size = size

                # Build FUNCTIONn register value.
                value = self.DATAVSIZE_MAP[size] | self.DWT_ACTION_DEBUG_EVENT | watch.func

                self.ap.write_memory(watch.comp_register_addr, addr)
                self.ap.write_memory(watch.comp_register_addr + self.DWT_FUNCTION_OFFSET, value)
                self.watchpoint_used += 1
                return True

        LOG.error('No more watchpoints are available, dropped watchpoint at 0x%08x', addr)
        return False






# pyOCD debugger
# Copyright (c) 2015-2019 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading
import socket

from .sockets import ListenerSocket
from .compatibility import to_bytes_safe

LOG = logging.getLogger(__name__)

class StreamServer(threading.Thread):
    """@brief File-like object that serves data over a TCP socket.

    The user can connect to the socket with telnet or netcat.

    The server thread will automatically be started by the constructor. To shut down the
    server and its thread, call the stop() method.
    """

    def __init__(self, port, serve_local_only=True, name=None, is_read_only=True, extra_info=None):
        """@brief Constructor.

        Starts the server immediately.

        @param self
        @param port The TCP/IP port number on which the server should listen. If 0 is passed,
            then an arbitrary unused port is selected by the OS. In this case, the `port` property
            can be used to get the actual port number.
        @param serve_local_only Whether to allow connections from remote clients.
        @param name Optional server name.
        @param is_read_only If the server is read-only, from the perspective of the client,
            then any incoming data sent by the client is discarded. Otherwise it is buffered so
            it can be read with the read() methods.
        @param extra_info Optional string with extra information about the server, e.g. "core 0".
        """
        super(StreamServer, self).__init__()
        self.name = name
        self._name = name
        self._extra_info = extra_info
        self._formatted_name = (name + " ") if (name is not None) else ""
        self._is_read_only = is_read_only
        self._abstract_socket = None
        self._abstract_socket = ListenerSocket(port, 4096)
        if not serve_local_only:
            # We really should be binding to explicit interfaces, not all available.
            self._abstract_socket.host = ''
        self._abstract_socket.init()
        self._port = self._abstract_socket.port
        self._buffer = bytearray()
        self._buffer_lock = threading.Lock()
        self.connected = None
        self._shutdown_event = threading.Event()
        self._is_running: bool = False
        self.daemon = True
        self.start()

    @property
    def port(self):
        return self._port

    @property
    def is_running(self) -> bool:
        return self._is_running

    def stop(self):
        self._shutdown_event.set()
        self.join()

    def run(self):
        LOG.info("%sserver started on port %d%s", self._formatted_name, self._port,
            (" (%s)" % self._extra_info) if self._extra_info else "")
        self.connected = None
        try:
            self._is_running = True
            while not self._shutdown_event.is_set():
                # Wait for a client to connect.
                # TODO support multiple client connections
                while not self._shutdown_event.is_set():
                    self.connected = self._abstract_socket.connect()
                    if self.connected is not None:
                        LOG.debug("%sclient connected", self._formatted_name)
                        break

                if self._shutdown_event.is_set():
                    break

                # Set timeout on new connection.
                self._abstract_socket.set_timeout(0.1)

                # Keep reading from the client until we either get a shutdown event, or
                # the client disconnects. The incoming data is appended to our read buffer.
                while not self._shutdown_event.is_set():
                    try:
                        data = self._abstract_socket.read()
                        if len(data) == 0:
                            # Client disconnected.
                            self._abstract_socket.close()
                            self.connected = None
                            break

                        if not self._is_read_only:
                            self._buffer_lock.acquire()
                            self._buffer += bytearray(data)
                            self._buffer_lock.release()
                    except socket.timeout:
                        pass
        finally:
            self._is_running = False
            self._abstract_socket.cleanup()
        LOG.info("%sserver stopped", self._formatted_name)

    def write(self, data):
        """@brief Write bytes into the connection."""
        # If nobody is connected, act like all data was written anyway.
        if self.connected is None:
            return 0
        data = to_bytes_safe(data)
        size = len(data)
        remaining = size
        while remaining:
            count = self._abstract_socket.write(data)
            remaining -= count
            if remaining:
                data = data[count:]
        return size

    def _get_input(self, length=-1):
        """@brief Extract requested amount of data from the read buffer."""
        self._buffer_lock.acquire()
        try:
            if length == -1:
                actualLength = len(self._buffer)
            else:
                actualLength = min(length, len(self._buffer))
            if actualLength:
                data = self._buffer[:actualLength]
                self._buffer = self._buffer[actualLength:]
            else:
                data = bytearray()
            return data
        finally:
            self._buffer_lock.release()

    def read(self, size=-1):
        """@brief Return bytes read from the connection."""
        if self.connected is None:
            return None

        # Extract requested amount of data from the read buffer.
        data = self._get_input(size)

        return data

    def readinto(self, b):
        """@brief Read bytes into a mutable buffer."""
        if self.connected is None:
            return None

        # Extract requested amount of data from the read buffer.
        b[:] = self._get_input()

        if len(b):
            return len(b)
        else:
            return None




# pyOCD debugger
# Copyright (c) 2017-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from time import (time, sleep)
from typing import (Any, Optional, TYPE_CHECKING)

if TYPE_CHECKING:
    from types import TracebackType

class Timeout:
    """@brief Timeout helper context manager.

    The recommended way to use this class is demonstrated here. It uses an else block on a
    while loop to handle the timeout. The code in the while loop must use a break statement
    to exit in the successful case.

    @code
    with Timeout(5, sleeptime=0.1) as t_o:
        while t_o.check(): # or "while not t_o.did_time_out"
            # Perform some operation, check, etc.
            if foobar:
                break
        else:
            print("Timed out!")
    @endcode

    Another method of using the class is to check the `did_time_out` property from within the
    while loop, as shown below.

    @code
    with Timeout(5) as t_o:
        while perform_some_test():
            # Check for timeout.
            if t_o.did_time_out:
                print("Timed out!")
                break
            sleep(0.1)
    @endcode

    You may also combine the call to check() in the while loop with other boolean expressions
    related to the operation being performed.

    If you pass a non-zero value for _sleeptime_ to the constructor, the check() method will
    automatically sleep by default starting with the second call. You can disable auto-sleep
    by passing `autosleep=False` to check().

    Passing a timeout of None to the constructor is allowed. In this case, check() will always return
    True and the loop must be exited via some other means.
    """

    def __init__(self, timeout: Optional[float], sleeptime: float = 0) -> None:
        """@brief Constructor.
        @param self
        @param timeout The timeout in seconds. May be None to indicate no timeout.
        @param sleeptime Time in seconds to sleep during calls to check(). Defaults to 0, thus
            check() will not sleep unless you pass a different value.
        """
        self._sleeptime = sleeptime
        self._timeout = timeout
        self._timed_out = False
        self._start = -1.0
        self._is_first_check = True
        self._is_running = False

    def __enter__(self) -> "Timeout":
        self.start()
        return self

    def __exit__(self, exc_type: type, value: Any, traceback: "TracebackType") -> bool:
        return False

    def start(self) -> None:
        """@brief Start or restart the timeout timer.

        This has precisely the same effect as entering `self` when used as a context manager.

        If called after the timeout has already been started, the effect is to reset the timeout from the
        current time.
        """
        self._is_running = True
        self._start = time()
        self._timed_out = False
        self._is_first_check = True

    def clear(self) -> None:
        """@brief Reset the timeout back to initial, non-running state.

        The timeout can be made to run again by calling start().
        """
        self._is_running = False
        self._timed_out = False
        self._is_first_check = True

    def check(self, autosleep: bool = True) -> bool:
        """@brief Check for timeout and possibly sleep.

        Starting with the second call to this method, it will automatically sleep before returning
        if:
            - The timeout has not yet occurred.
            - A non-zero _sleeptime_ was passed to the constructor.
            - The _autosleep_ parameter is True.

        This method is intended to be used as the predicate of a while loop.

        If this method is called prior to the timeout being started (by the start() method or entering
        it as a context manager) this the return value will always be True (not timeed out). Only after
        the timeout is running will the elapsed time be tested.

        @param self
        @param autosleep Whether to sleep if not timed out yet. The sleeptime passed to the
            constructor must have been non-zero.
        @retval True The timeout has _not_ occurred.
        @retval False Timeout is passed and the loop should be exited.
        """
        # Check for a timeout.
        if self._is_running and (self._timeout is not None) and ((time() - self._start) > self._timeout):
            self._timed_out = True
        # Sleep if appropriate.
        elif (not self._is_first_check) and autosleep and self._sleeptime:
            sleep(self._sleeptime)
        self._is_first_check = False
        return not self._timed_out

    @property
    def is_running(self) -> bool:
        """@brief Whether the timeout object has started timing."""
        return self._is_running

    @property
    def did_time_out(self) -> bool:
        """@brief Whether the timeout has occurred as of the time when this property is accessed."""
        self.check(autosleep=False)
        return self._timed_out




# pyOCD debugger
# Copyright (c) 2006-2019 Arm Limited
# Copyright (c) 2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import (Optional, TYPE_CHECKING)

from .board import Board
from .board_ids import (BoardInfo, BOARD_ID_TO_INFO)

if TYPE_CHECKING:
    from ..core.session import Session

LOG = logging.getLogger(__name__)

class MbedBoard(Board):
    """@brief Mbed board class.

    This class used to implement the lookup of board name and other info based on the board ID contained
    in the probe's serial number string. With CMSIS-DAP v2.1 now having support for reporting the board
    and target info, the logic is much more complex and is mostly performed in CMSISDAPProbe. This
    class now simply verifies that the given board_id is known and logs a warning if not.

    If the board ID is all "0" characters, it indicates the firmware is generic and doesn't have an
    associated board.
    """
    def __init__(self,
            session: "Session",
            target: Optional[str] = None,
            board_info: Optional["BoardInfo"] = None,
            board_id: Optional[str] = None,
            ) -> None:
        """@brief Constructor.

        Validates the given board_id, if any.
        """
        # Check for an all-zero board ID. This indicates a standalone probe or generic firmware.
        if board_id == "0000":
            pass
        elif board_id:
            # Attempt to look up the board ID in our table.
            try:
                info_from_table = BOARD_ID_TO_INFO[board_id]
                if not board_info:
                    board_info = info_from_table
            except KeyError:
                LOG.warning("Board ID %s is not recognized", board_id)

                # If we don't have board info, then construct one indicating the board is unknown.
                if not board_info:
                    board_info = BoardInfo("Unknown Board")

        self._board_id = board_id

        super().__init__(session, target, board_info)

    @property
    def board_id(self) -> Optional[str]:
        return self._board_id
